#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the KKMC WorkflowModule."""

from __future__ import print_function
from __future__ import absolute_import
import unittest
import os
import os.path
import shutil
import tempfile
from mock import patch, MagicMock as Mock, mock_open

from parameterized import parameterized, param

from DIRAC import gLogger, S_OK, S_ERROR
from ILCDIRAC.Workflow.Modules.KKMCAnalysis import KKMCAnalysis
from Tests.Utilities.GeneralUtils import assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.KKMCAnalysis'
MODULEBASE_NAME = 'ILCDIRAC.Workflow.Modules.ModuleBase'
PROXYINFO_NAME = 'DIRAC.Core.Security.ProxyInfo'
# pylint: disable=too-many-public-methods, protected-access

gLogger.setLevel("ERROR")
gLogger.showHeaders(True)


def cleanup(tempdir):
  """Remove files after run."""
  try:
    shutil.rmtree(tempdir)
  except OSError:
    pass


@patch("%s.getProxyInfoAsString" % MODULEBASE_NAME, new=Mock(return_value=S_OK()))
@patch("%s.getProxyInfoAsString" % PROXYINFO_NAME, new=Mock(return_value=S_OK()))
class TestKKMCAnalysis(unittest.TestCase):
  """test KKMCAnalysis."""

  def assertIn(self, *args, **kwargs):
    """make this existing to placate pylint."""
    return super(TestKKMCAnalysis, self).assertIn(*args, **kwargs)

  @patch("%s.getProxyInfoAsString" % MODULEBASE_NAME, new=Mock(return_value=S_OK()))
  @patch("%s.getProxyInfoAsString" % PROXYINFO_NAME, new=Mock(return_value=S_OK()))
  def setUp(self):
    self.kkmc = KKMCAnalysis()
    self.curdir = os.getcwd()
    self.tempdir = tempfile.mkdtemp("", dir="./")
    os.chdir(self.tempdir)
    self.kkmc.ops = Mock()

  def tearDown(self):
    os.chdir(self.curdir)
    cleanup(self.tempdir)


class TestKKMCAnalysisRunit(TestKKMCAnalysis):
  """test KKMC runtIt."""

  def setUp(self):
    super(TestKKMCAnalysisRunit, self).setUp()
    self.logFileName = "localEnv.log"
    with open(self.logFileName, "w") as logF:
      logF.write("logged the logging logs")

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_KKMC_runIt_success(self):
    """KKMC.runit ................................................................................."""
    self.kkmc.platform = 'Windows'
    self.kkmc.applicationLog = self.logFileName
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, True, True])):
      res = self.kkmc.runIt()
    print(res)
    assertDiracSucceeds(res, self)

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_KKMC_runIt_failure_LogFile(self):
    """KKMC.runit failure with applicationLog......................................................"""
    self.kkmc.platform = "Windows"
    self.kkmc.applicationLog = self.logFileName
    self.kkmc.ignoreapperrors = False
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.kkmc.runIt()
    self.assertIn("did not produce the expected log", res['Message'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_KKMC_runIt_failure_LogFile_ignore(self):
    """KKMC.runit failure with applicationLog but ignore..........................................."""
    self.kkmc.platform = "Windows"
    self.kkmc.applicationLog = self.logFileName
    self.kkmc.ignoreapperrors = True
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.kkmc.runIt()
    assertDiracSucceeds(res, self)

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_KKMC_runIt_failure_NoLogFile(self):
    """KKMC.runit failure with applicationLog not set............................................."""
    self.kkmc.platform = "Windows"
    self.kkmc.ignoreapperrors = True
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.kkmc.runIt()
    self.assertIn("No Log file provide", res['Message'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_KKMC_runIt_failure_NoPlatform(self):
    """KKMC.runit failure with platform ........................................................."""
    self.kkmc.applicationLog = self.logFileName
    self.kkmc.ignoreapperrors = True
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[False, False, False, False])):
      res = self.kkmc.runIt()
    self.assertIn("No ILC platform selected", res['Message'])

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh")))
  @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0, "AllGood"))))
  def test_KKMC_runIt_success_LogAndScriptPresent(self):
    """KKMC.runit success log and script exist..................................................."""
    self.kkmc.platform = "Windows"
    self.kkmc.applicationLog = self.logFileName
    self.kkmc.ignoreapperrors = True
    with open("kkmc__Run_.sh", "w") as scr:
      scr.write("content")
    with open("KKMC__Steer_.input", "w") as scr:
      scr.write("content")
    with open(self.logFileName, "w") as scr:
      scr.write("content")
    # side effect for Script, userlibs, log, logAfter
    with patch("os.path.exists", new=Mock(side_effect=[True, True, False, True])):
      res = self.kkmc.runIt()
    assertDiracSucceeds(res, self)

  # @parameterized.expand([('NumberOfEvents', 100, 'n_events = 100'),
  #                        ('OutputFile', 'test.slcio', 'sample_format = lcio'),
  #                        ('OutputFile', 'test.ascii', 'sample_format = ascii'),
  #                        ('OutputFile', 'test.stdhep', 'sample_format = stdhep'),
  #                        ('randomSeed', '321', 'seed = 321'),
  #                        param('KKMCRawSin', True, 'sample_format', pUnExpected=True),
  #                        ])
  # @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_OK("setup.sh") ) )
  # @patch("%s.shellCall" % MODULE_NAME, new=Mock(return_value=S_OK((0,"AllGood")) ) )
  # def test_KKMC_runIt_success_configFile(self, pName, pValue, pExpected, pUnExpected=None):
  #   """kkmc.runit success with configFile........................................................."""
  #   self.kkmc.platform = "Windows"
  #   self.kkmc.applicationLog = self.logFileName
  #   self.kkmc.kkmcConfigFile  = "kkmc instructions"
  #   setattr(self.kkmc, pName, pValue)
  #   ## side effect for Steering1, Steering2, Script, userlib, log, logAfter
  #   with patch("os.path.exists", new=Mock(side_effect=[False, False, False, True] ) ):
  #     res = self.kkmc.runIt()
  #   assertDiracSucceeds( res, self )
  #   self.assertEqual( self.kkmc.kkmcConfigFile , "kkmc instructions" )
  #   self.assertIn( "kkmc instructions", open("KKMC__Steer_.input").read())
  #   if not pUnExpected:
  #     self.assertIn(pExpected, open("KKMC__Steer_.input").read())
  #   else:
  #     self.assertNotIn(pExpected, open("KKMC__Steer_.input").read())

  @patch("%s.getEnvironmentScript" % MODULE_NAME, new=Mock(return_value=S_ERROR("missing setup.sh")))
  def test_KKMC_runIt_fail_env(self):
    """kkmc.runit failed to get env................................................................"""
    self.kkmc.platform = "Windows"
    self.kkmc.applicationLog = self.logFileName
    res = self.kkmc.runIt()
    self.assertEqual(res['Message'], "missing setup.sh")


class TestKKMCAnalysisASI(TestKKMCAnalysis):
  """kkmc.ApplicationSpecificInputs."""

  @patch.dict(os.environ, {"JOBID": "12345"})
  def test_KKMC_ASI_NoVariables(self):
    """kkmc.applicationSpecificInputs: checks that no variables have been set after this call......"""
    gLogger.setLevel("ERROR")
    self.kkmc.workflow_commons = dict()
    self.kkmc.applicationSpecificInputs()
    self.assertFalse(self.kkmc.jobReport or self.kkmc.productionID)


def runTests():
  """Runs our tests."""
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestKKMCAnalysis)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestKKMCAnalysisRunit))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestKKMCAnalysisASI))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)


if __name__ == '__main__':
  runTests()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Module to upload specified job output files according to the parameters defined in the user workflow.

:author: S. Poss
:since: Sep 01, 2010
"""

from __future__ import absolute_import
import os
import random
import time

from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import List
import DIRAC

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.ProductionData import constructUserLFNs
from ILCDIRAC.Core.Utilities.ResolveSE import getDestinationSEList
from ILCDIRAC.Core.Utilities.Splitting import addJobIndexToFilename

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class UserJobFinalization(ModuleBase):
  """User Job finalization: takes care of uploading the output data to the specified storage elements.

  If it does not work, it will upload to a Failover SE, then register the request to replicate and remove.
  """
  #############################################################################

  def __init__(self):
    """Module initialization."""
    super(UserJobFinalization, self).__init__()
    self.version = __RCSID__
    self.enable = True
    self.defaultOutputSE = gConfig.getValue('/Resources/StorageElementGroups/Tier1-USER', [])
    self.failoverSEs = gConfig.getValue('/Resources/StorageElementGroups/Tier1-Failover', [])
    # List all parameters here
    self.userFileCatalog = self.ops.getValue('/UserJobs/Catalogs', ['FileCatalog'])
    # Always allow any files specified by users
    self.outputDataFileMask = ''
    self.userOutputData = ''
    self.userOutputSE = ''
    self.userOutputPath = ''
    self.jobReport = None

  #############################################################################
  def applicationSpecificInputs(self):
    """By convention the module parameters are resolved here."""

    # Earlier modules may have populated the report objects
    self.jobReport = self.workflow_commons.get('JobReport', self.jobReport)

    self.enable = self.step_commons.get('Enable', self.enable)
    if not isinstance(self.enable, bool):
      LOG.warn('Enable flag set to non-boolean value %s, setting to False' % self.enable)
      self.enable = False

    self.jobID = os.environ.get('JOBID', self.jobID)
    if self.jobID != 0:
      LOG.verbose('Found WMS JobID = %s' % self.jobID)
    else:
      LOG.info('No WMS JobID found, disabling module via control flag')
      self.enable = False

    # Use LHCb utility for local running via dirac-jobexec
    self.userOutputData = self.workflow_commons.get('UserOutputData', self.userOutputData)
    if self.userOutputData and not isinstance(self.userOutputData, list):
      self.userOutputData = [i.strip() for i in self.userOutputData.split(';')]

    specifiedSE = self.workflow_commons.get('UserOutputSE', '')
    if not isinstance(specifiedSE, list) and specifiedSE:
      self.userOutputSE = [i.strip() for i in specifiedSE.split(';')]
      LOG.debug("userOutputSE is: " + str(self.userOutputSE))
    else:
      LOG.verbose('No UserOutputSE specified, using default value: %s' % (', '.join(self.defaultOutputSE)))
      self.userOutputSE = self.defaultOutputSE

    self.userOutputPath = self.workflow_commons.get('UserOutputPath', self.userOutputPath)

    return S_OK('Parameters resolved')

  #############################################################################
  def execute(self):
    """Main execution function."""
    # Have to work out if the module is part of the last step i.e.
    # user jobs can have any number of steps and we only want
    # to run the finalization once. Not a problem if this is not the last step so return S_OK()
    resultLS = self.isLastStep()
    if not resultLS['OK']:
      return S_OK()

    self.logWorkingDirectory()

    resultIV = self.resolveInputVariables()
    if not resultIV['OK']:
      LOG.error("Failed to resolve input parameters:", resultIV['Message'])
      return resultIV

    LOG.info('Initializing %s' % self.version)
    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'],
                                                                   self.stepStatus['OK']))
      return S_OK('No output data upload attempted')

    if not self.userOutputData:
      LOG.info('No user output data is specified for this job, nothing to do')
      return S_OK('No output data to upload')

    # Determine the final list of possible output files for the
    # workflow and all the parameters needed to upload them.
    outputList = self.getOutputList()

    userOutputLFNs = []
    if self.userOutputData:
      resultOLfn = self.constructOutputLFNs()
      if not resultOLfn['OK']:
        LOG.error('Could not create user LFNs', resultOLfn['Message'])
        return resultOLfn
      userOutputLFNs = resultOLfn['Value']

    LOG.verbose('Calling getCandidateFiles( %s, %s, %s)' % (outputList, userOutputLFNs, self.outputDataFileMask))
    LOG.debug("IgnoreAppErrors? '%s' " % self.ignoreapperrors)
    resultCF = self.getCandidateFiles(outputList, userOutputLFNs, self.outputDataFileMask)
    if not resultCF['OK']:
      if not self.ignoreapperrors:
        LOG.error(resultCF['Message'])
        self.setApplicationStatus(resultCF['Message'])
        return S_OK()
    fileDict = resultCF['Value']

    resultFMD = self.getFileMetadata(fileDict)
    if not resultFMD['OK']:
      if not self.ignoreapperrors:
        LOG.error(resultFMD['Message'])
        self.setApplicationStatus(resultFMD['Message'])
        return S_OK()

    if not resultFMD['Value']:
      if not self.ignoreapperrors:
        LOG.info('No output data files were determined to be uploaded for this workflow')
        self.setApplicationStatus('No Output Data Files To Upload')
        return S_OK()

    fileMetadata = resultFMD['Value']

    # First get the local (or assigned) SE to try first for upload and others in random fashion
    resultSEL = getDestinationSEList('Tier1-USER', DIRAC.siteName(), outputmode='local')
    if not resultSEL['OK']:
      LOG.error('Could not resolve output data SE', resultSEL['Message'])
      self.setApplicationStatus('Failed To Resolve OutputSE')
      return resultSEL
    localSE = resultSEL['Value']

    orderedSEs = [se for se in self.defaultOutputSE if se not in localSE and se not in self.userOutputSE]

    orderedSEs = localSE + List.randomize(orderedSEs)
    if self.userOutputSE:
      prependSEs = []
      for userSE in self.userOutputSE:
        if userSE not in orderedSEs:
          prependSEs.append(userSE)
      orderedSEs = prependSEs + orderedSEs

    LOG.info('Ordered list of output SEs is: %s' % (', '.join(orderedSEs)))
    final = {}
    for fileName, metadata in fileMetadata.items():
      final[fileName] = metadata
      final[fileName]['resolvedSE'] = orderedSEs

    self.injectJobIndex(final)

    # At this point can exit and see exactly what the module will upload
    self.printOutputInfo(final)
    if not self.enable:
      return S_OK('Module is disabled by control flag')

    # Instantiate the failover transfer client with the global request object
    failoverTransfer = FailoverTransfer(self._getRequestContainer())

    # One by one upload the files with failover if necessary
    filesToReplicate = {}
    filesToFailover = {}
    filesUploaded = []
    self.transferAndRegisterFiles(final, failoverTransfer, filesToFailover, filesUploaded, filesToReplicate)

    # if there are files to be failovered, we do it now
    resultTRFF = self.transferRegisterAndFailoverFiles(failoverTransfer, filesToFailover, filesUploaded)
    cleanUp = resultTRFF['Value']['cleanUp']

    # For files correctly uploaded must report LFNs to job parameters
    if filesUploaded:
      report = ', '.join(filesUploaded)
      self.jobReport.setJobParameter('UploadedOutputData', report)

    self.workflow_commons['Request'] = failoverTransfer.request

    # If some or all of the files failed to be saved to failover
    if cleanUp:
      # Leave any uploaded files just in case it is useful for the user
      # do not try to replicate any files.
      return S_ERROR('Failed To Upload Output Data')

    # If there is now at least one replica for uploaded files can trigger replication
    datMan = DataManager(catalogs=self.userFileCatalog)
    LOG.info('Sleeping for 10 seconds before attempting replication of recently uploaded files')
    time.sleep(10)
    for lfn, repSE in filesToReplicate.items():
      resultRAR = datMan.replicateAndRegister(lfn, repSE)
      if not resultRAR['OK']:
        LOG.info('Replication failed with below error but file already exists in Grid storage with \
        at least one replica:\n%s' % (resultRAR))

    self.generateFailoverFile()

    self.setApplicationStatus('Job Finished Successfully')
    return S_OK('Output data uploaded')

  #############################################################################
  @staticmethod
  def getCurrentOwner():
    """Simple function to return current DIRAC username."""
    result = getProxyInfo()
    if not result['OK']:
      return S_ERROR('Could not obtain proxy information')

    if 'username' not in result['Value']:
      return S_ERROR('Could not get username from proxy')

    username = result['Value']['username']
    return S_OK(username)
  #############################################################################

  @staticmethod
  def getCurrentVO():
    """Simple function to return current DIRAC username."""
    result = getProxyInfo()
    if not result['OK']:
      return S_ERROR('Could not obtain proxy information')

    if 'group' not in result['Value']:
      return S_ERROR('Could not get group from proxy')

    group = result['Value']['group']
    vo = group.split("_")[0]
    return S_OK(vo)

  def constructOutputLFNs(self):
    """Returns a list of the outputDataLFNs."""
    LOG.info('Constructing user output LFN(s) for %s' % (', '.join(self.userOutputData)))
    if not self.jobID:
      self.jobID = 12345
    owner = self.workflow_commons.get('Owner', '')
    if not owner:
      res = self.getCurrentOwner()
      if not res['OK']:
        LOG.error('Could not find proxy')
        return S_ERROR('Could not obtain owner from proxy')
      owner = res['Value']

    vo = self.workflow_commons.get('VO', '')
    if not vo:
      res = self.getCurrentVO()
      if not res['OK']:
        LOG.error('Failed finding the VO')
        return S_ERROR('Could not obtain VO from proxy')
      vo = res['Value']

    result = constructUserLFNs(int(self.jobID), vo, owner, self.userOutputData, self.userOutputPath)
    if not result['OK']:
      LOG.error('Could not create user LFNs', result['Message'])
      return result

    return result

  def isLastStep(self):
    """returns S_OK() if this is the last step."""
    currentStep = int(self.step_commons['STEP_NUMBER'])
    totalSteps = int(self.workflow_commons['TotalSteps'])
    if currentStep == totalSteps:
      LOG.verbose("This is the last step, let's finalize this userjob")
      return S_OK()
    else:
      LOG.verbose('Current step = %s, total steps of workflow = %s, UserJobFinalization will enable itself only \
      at the last workflow step.' % (currentStep, totalSteps))
      return S_ERROR("Not the last step")

  def getOutputList(self):
    """returns list of dictionary with output files, paths and SEs userOutputData is list of files specified by user.

    Q: Why is this taking the last part of the filename as outputPath????
    A: outputPath is not actually used anywhere
    """
    outputList = []
    for filename in self.userOutputData:
      outputList.append({'outputPath': filename.split('.')[-1].upper(),
                         'outputDataSE': self.userOutputSE,
                         'outputFile': os.path.basename(filename)})
    LOG.debug("OutputList: %s" % outputList)
    return outputList

  def printOutputInfo(self, final):
    """print some information about what would be uploaded."""
    if not self.enable:
      LOG.info('Module is disabled by control flag, would have attempted to upload the following files')
    else:
      LOG.info('Attempt to upload the following files')
    for fileName, metadata in final.items():
      LOG.info('--------%s--------' % fileName)
      for metaName, metaValue in metadata.items():
        LOG.info('%s = %s' % (metaName, metaValue))

  def injectJobIndex(self, final):
    """add the jobIndex to the output file name."""
    if self.workflow_commons.get('JobIndex') is None:
      return
    if self.workflow_commons.get('DoNotAlterOutputData'):
      return

    jobIndex = self.workflow_commons.get('JobIndex')
    for _, metaInfo in final.items():
      orgLFN = metaInfo['lfn']
      metaInfo['lfn'] = addJobIndexToFilename(orgLFN, jobIndex)
      LOG.info("Changing '%s' to '%s'" % (orgLFN, metaInfo['lfn']))

  def transferAndRegisterFiles(self, final, failoverTransfer, filesToFailover, filesUploaded, filesToReplicate):
    """transfer and register files to storage elements.

    fills filesToFailover, filesUploaded and filesToReplicate dicts
    """

    for fileName, metadata in final.items():
      LOG.info("Attempting to store file %s to the following SE(s):\n%s" % (fileName,
                                                                                 ', '.join(metadata['resolvedSE'])))
      resultFT = failoverTransfer.transferAndRegisterFile(fileName,
                                                          metadata['localpath'],
                                                          metadata['lfn'],
                                                          metadata['resolvedSE'],
                                                          fileMetaDict=metadata['filedict'],
                                                          fileCatalog=self.userFileCatalog)
      if not resultFT['OK']:
        LOG.error('Could not transfer and register %s with metadata:\n %s' % (fileName, metadata))
        filesToFailover[fileName] = metadata
      else:
        # Only attempt replication after successful upload and if there is more than one userOutputSE
        lfn = metadata['lfn']
        filesUploaded.append(lfn)
        seList = self.userOutputSE
        replicateSE = ''
        uploadedSE = resultFT['Value'].get('uploadedSE', '')
        if uploadedSE:
          for se in seList:
            if se != uploadedSE:
              replicateSE = se
              break

        if replicateSE and lfn:
          LOG.info('Will attempt to replicate %s to %s' % (lfn, replicateSE))
          filesToReplicate[lfn] = replicateSE

  def transferRegisterAndFailoverFiles(self, failoverTransfer, filesToFailover, filesUploaded):
    """transfer and failover request."""
    cleanUp = False
    for fileName, metadata in filesToFailover.items():
      random.shuffle(self.failoverSEs)
      targetSE = metadata['resolvedSE'][0]

      # make sure we don't upload to one SE and then try to move it there again
      failoverSEs = list(self.failoverSEs)
      if targetSE in failoverSEs:
        failoverSEs.remove(targetSE)
        if not failoverSEs:
          LOG.error("No more failoverSEs to consider, skipping file: %s" % fileName)
          LOG.error("TargetSE: %s, All FailoverSEs: %s" % (targetSE, self.failoverSEs))
          cleanUp = True
          continue
      LOG.verbose("TargetSE: %s, All FailoverSEs: %s, Cleaned FailoverSEs: %s"
                       % (targetSE, self.failoverSEs, failoverSEs))

      metadata['resolvedSE'] = failoverSEs
      resultFT = failoverTransfer.transferAndRegisterFileFailover(fileName,
                                                                  metadata['localpath'],
                                                                  metadata['lfn'],
                                                                  targetSE,
                                                                  failoverSEs,
                                                                  fileMetaDict=metadata['filedict'],
                                                                  fileCatalog=self.userFileCatalog)
      if not resultFT['OK']:
        LOG.error('Could not transfer and register %s with metadata:\n %s' % (fileName, metadata))
        cleanUp = True
        continue  # for users can continue even if one completely fails
      else:
        lfn = metadata['lfn']
        filesUploaded.append(lfn)

    return S_OK(dict(cleanUp=cleanUp))
#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for the UploadLogFile module."""

from __future__ import print_function
from __future__ import absolute_import
import sys
import unittest
from mock import patch, MagicMock as Mock

from Tests.Utilities.GeneralUtils import assertDiracFails, assertDiracSucceeds, \
    assertDiracSucceedsWith_equals, assertMockCalls
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.UploadLogFile'

# pylint: disable=protected-access


class UploadLogFileTestCase(unittest.TestCase):
  """Contains tests for the UploadLogFile class."""

  ops_dict = {'/LogFiles/CLIC/Extensions':
               ['*.txt', '*.log', '*.out', '*.output', '*.xml', '*.sh', '*.info', '*.err', '*.root']}

  def setUp(self):
    """set up the objects."""
    # Mock out modules that spawn other threads
    mocked_modules = {'DIRAC.DataManagementSystem.Client.DataManager': Mock()}
    self.log_mock = Mock()
    self.patches = [patch('%s.LOG' % MODULE_NAME, new=self.log_mock),
                    patch.dict(sys.modules, mocked_modules)]
    for patcher in self.patches:
      patcher.start()

    from ILCDIRAC.Workflow.Modules.UploadLogFile import UploadLogFile
    self.ulf = UploadLogFile()
    self.ulf.jobID = 8194
    self.ulf.workflow_commons = {'Request': 'something'}
    ops_mock = Mock()
    ops_mock.getValue.side_effect = lambda key, _: UploadLogFileTestCase.ops_dict[key]
    self.ulf.ops = ops_mock

  def tearDown(self):
    """Clean up test resources."""
    for patcher in self.patches:
      patcher.stop()

  def test_execute(self):
    stat_list = [('', '', '', '', '', '', 148), ('', '', '', '', '', '', 2984828952984), OSError('mock_oserr')]
    glob_list = [['ignore_me', 'file_1', 'file_2', 'file_3'], [], [], [], [], [], [], [], [], []]
    UploadLogFileTestCase.ops_dict['/LogFiles/CLIC/Extensions'] = []
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_ERROR('my_testerr'))), \
         patch('%s.os.stat' % MODULE_NAME, new=Mock(side_effect=stat_list)), \
         patch('%s.glob.glob' % MODULE_NAME, new=Mock(side_effect=glob_list)), \
         patch('%s.os.path.isfile' % MODULE_NAME, new=Mock(side_effect=[False, True, True, True])), \
         patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      # FIXME: Deactivated since fails in some environments
      # assertEqualsImproved( log_mock.error.mock_calls, [
      #  call('Failed to resolve input parameters:', 'my_testerr'),
      #  call('Log file found to be greater than maximum of %s bytes' % self.ulf.logSizeLimit, 'file_2'),
      #  call('Completely failed to select relevant log files.', 'Could not determine log files') ], self )

  def test_execute_nologs(self):
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
        patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)), \
        patch('%s.os.makedirs' % MODULE_NAME, new=Mock(side_effect=OSError('os_mkdir_failed_testerr_populate'))), \
        patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      self.log_mock.error.assert_called_once_with('Completely failed to populate temporary log file directory.', '')

  def test_execute_disabled(self):
    self.ulf.enable = False
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
         patch.object(self.ulf, '_populateLogDirectory', new=Mock(return_value=S_OK())), \
         patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceedsWith_equals(self.ulf.execute(), 'Module is disabled by control flag', self)
      self.assertFalse(self.log_mock.error.called)

  def test_execute_tarfails(self):
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
        patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
        patch.object(self.ulf, '_populateLogDirectory', new=Mock(return_value=S_OK())), \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['file1', 'testfile2'])), \
        patch('%s.os.path.islink' % MODULE_NAME, new=Mock(side_effect=[True, False])), \
        patch('%s.os.chmod' % MODULE_NAME, new=Mock(side_effect=OSError('chmod_mock_testerr'))), \
        patch('%s.os.path.realpath' % MODULE_NAME, new=Mock(return_value='./job/log/prodID/jobID')), \
        patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/my/inval/cwd')), \
        patch('%s.os.chdir' % MODULE_NAME), \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['mytarfile', 'otherfile'])), \
        patch('%s.tarfile.open' % MODULE_NAME), \
        patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)), \
        patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      assertMockCalls(self.log_mock.error, [
          'Could not set permissions of log files to 0755 with message:\nchmod_mock_testerr',
          ('Failed to create tar file from directory', './job/log/prodID/jobID File was not created'),
          ('Problem changing shared area permissions', 'chmod_mock_testerr')], self)

  def test_execute_all_works(self):
    self.ulf.failoverTest = False
    se_mock = Mock()
    self.ulf.logSE = se_mock
    se_mock.putFile.return_value = S_OK({'Failed': [], 'Successful': []})
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
         patch.object(self.ulf, '_populateLogDirectory', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_setLogFilePermissions', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_tarTheLogFiles', new=Mock(return_value=S_OK({'fileName': 'some_name'}))), \
         patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      self.assertFalse(self.log_mock.error.called)

  def test_execute_failover_fails(self):
    self.ulf.failoverTest = False
    se_mock = Mock()
    se_mock.name = 'mySEMOCK'
    self.ulf.logSE = se_mock
    se_mock.putFile.return_value = S_OK({'Failed': ['some_file_failed'], 'Successful': []})
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
         patch.object(self.ulf, '_populateLogDirectory', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_setLogFilePermissions', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_tarTheLogFiles', new=Mock(return_value=S_OK({'fileName': 'some_name'}))), \
         patch.object(self.ulf, '_tryFailoverTransfer', new=Mock(return_value=S_OK())), \
         patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      self.log_mock.error.assert_called_once_with(
          "Completely failed to upload log files to mySEMOCK, will attempt upload to failover SE",
          {'Successful': [], 'Failed': ['some_file_failed']})

  def test_execute_failover_works(self):
    self.ulf.failoverTest = False
    se_mock = Mock()
    se_mock.name = 'mySEMOCK'
    self.ulf.logSE = se_mock
    se_mock.putFile.return_value = S_OK({'Failed': ['some_file_failed'], 'Successful': []})
    request_mock = Mock()
    request_mock.RequestName = 'mymockreq'
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
         patch.object(self.ulf, '_populateLogDirectory', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_setLogFilePermissions', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_tarTheLogFiles', new=Mock(return_value=S_OK({'fileName': 'some_name'}))), \
         patch.object(self.ulf, '_tryFailoverTransfer', new=Mock(return_value=S_OK({'Request': request_mock, 'uploadedSE': 'mock_se'}))), \
         patch.object(self.ulf, '_createLogUploadRequest', new=Mock(return_value=S_OK())) as uploadreq_mock, \
         patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      self.log_mock.error.assert_called_once_with(
          "Completely failed to upload log files to mySEMOCK, will attempt upload to failover SE",
          {'Successful': [], 'Failed': ['some_file_failed']})
      uploadreq_mock.assert_called_once_with('mySEMOCK', '', 'mock_se')

  def test_execute_logupload_fails(self):
    self.ulf.failoverTest = False
    se_mock = Mock()
    se_mock.name = 'mySEMOCK'
    self.ulf.logSE = se_mock
    se_mock.putFile.return_value = S_OK({'Failed': ['some_file_failed'], 'Successful': []})
    request_mock = Mock()
    request_mock.RequestName = 'mymockreq'
    with patch.object(self.ulf, 'resolveInputVariables', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_determineRelevantFiles', new=Mock(return_value=S_OK(['first_selected_file', 'other_files']))), \
         patch.object(self.ulf, '_populateLogDirectory', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_setLogFilePermissions', new=Mock(return_value=S_OK())), \
         patch.object(self.ulf, '_tarTheLogFiles', new=Mock(return_value=S_OK({'fileName': 'some_name'}))), \
         patch.object(self.ulf, '_tryFailoverTransfer', new=Mock(return_value=S_OK({'Request': request_mock, 'uploadedSE': 'mock_se'}))), \
         patch.object(self.ulf, '_createLogUploadRequest', new=Mock(return_value=S_ERROR('upload_mock_err'))) as uploadreq_mock, \
         patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracSucceeds(self.ulf.execute(), self)
      assertMockCalls(self.log_mock.error, [
          ('Completely failed to upload log files to mySEMOCK, will attempt upload to failover SE',
           {'Successful': [], 'Failed': ['some_file_failed']}),
          ('Failed to create failover request', 'upload_mock_err')], self)
      uploadreq_mock.assert_called_once_with('mySEMOCK', '', 'mock_se')

  def test_populatelogdir_nopermissions(self):
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=False)), \
        patch('%s.os.makedirs' % MODULE_NAME, new=Mock(return_value=True)), \
        patch('%s.os.chmod' % MODULE_NAME, new=Mock(side_effect=OSError('permission_denied_testerr'))), \
        patch('%s.shutil.copy' % MODULE_NAME, new=Mock(side_effect=OSError('shutil_mockerr'))), \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=[])), \
        patch('%s.UploadLogFile.logWorkingDirectory' % MODULE_NAME, new=Mock()):
      assertDiracFails(self.ulf._populateLogDirectory(['some_file']), self)
      self.log_mock.error.assert_called_once_with(
          'PopulateLogDir: Could not set logdir permissions to 0755:', ' (permission_denied_testerr)')
      self.log_mock.exception.assert_called_once_with(
          'PopulateLogDir: Exception while trying to copy file.', 'some_file', 'shutil_mockerr')


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(UploadLogFileTestCase)
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for OutputDataPolicy."""

from __future__ import absolute_import
from itertools import product
from parameterized import parameterized

WORKFLOW_XML = """
<?xml version="1.0" encoding="UTF-8" ?><Workflow>
<origin></origin>
<description><![CDATA[]]></description>
<descr_short></descr_short>
<version>0.0</version>
<type></type>
<name>ILD-Opt_sim_1000_6f_WWS_s5_v02_20190619_1</name>
<Parameter name="JobType" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified type"><value><![CDATA[%(JOB_TYPE)s]]></value></Parameter>
<Parameter name="Priority" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Priority"><value><![CDATA[1]]></value></Parameter>
<Parameter name="JobGroup" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified job group"><value><![CDATA[@{PRODUCTION_ID}]]></value></Parameter>
<Parameter name="JobName" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Name of Job"><value><![CDATA[Name]]></value></Parameter>
<Parameter name="Site" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Site Requirement"><value><![CDATA[ANY]]></value></Parameter>
<Parameter name="Origin" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Origin of client"><value><![CDATA[DIRAC]]></value></Parameter>
<Parameter name="StdOutput" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Standard output file"><value><![CDATA[std.out]]></value></Parameter>
<Parameter name="StdError" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Standard error file"><value><![CDATA[std.err]]></value></Parameter>
<Parameter name="InputData" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Default null input data value"><value><![CDATA[]]></value></Parameter>
<Parameter name="LogLevel" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified logging level"><value><![CDATA[verbose]]></value></Parameter>
<Parameter name="arguments" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Arguments to executable Step"><value><![CDATA[]]></value></Parameter>
<Parameter name="ParametricInputData" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Default null parametric input data value"><value><![CDATA[]]></value></Parameter>
<Parameter name="ParametricInputSandbox" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Default null parametric input sandbox value"><value><![CDATA[]]></value></Parameter>
<Parameter name="Platform" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Platform ( Operating System )"><value><![CDATA[x86_64-slc5-gcc43-opt]]></value></Parameter>
<Parameter name="IS_PROD" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="This job is a production job"><value><![CDATA[True]]></value></Parameter>
<Parameter name="MaxCPUTime" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="CPU time in secs"><value><![CDATA[300000]]></value></Parameter>
<Parameter name="CPUTime" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="CPU time in secs"><value><![CDATA[300000]]></value></Parameter>
<Parameter name="productionVersion" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdAPIVersion"><value><![CDATA[$Id$]]></value></Parameter>
<Parameter name="PRODUCTION_ID" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProductionID"><value><![CDATA[00012345]]></value></Parameter>
<Parameter name="JOB_ID" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProductionJobID"><value><![CDATA[00012345]]></value></Parameter>
<Parameter name="emailAddress" type="string" linked_module="" linked_parameter="" in="True" out="False" description="CrashEmailAddress"><value><![CDATA[ilcdirac-support@cern.ch]]></value></Parameter>
<Parameter name="SoftwarePackages" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="ILC Software Packages to be installed"><value><![CDATA[ildconfig.v02-00-02;ddsim.ILCSoft-02-00-02_gcc49]]></value></Parameter>
<Parameter name="ILDConfigPackage" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="ILDConfig package"><value><![CDATA[ILDConfigv02-00-02]]></value></Parameter>
<Parameter name="BannedSites" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="Site excluded by user"><value><![CDATA[]]></value></Parameter>
<Parameter name="Energy" type="float" linked_module="" linked_parameter="" in="True" out="False" description="Energy used"><value><![CDATA[1000.0]]></value></Parameter>
<Parameter name="TotalSteps" type="String" linked_module="" linked_parameter="" in="True" out="False" description="Total number of steps"><value><![CDATA[1]]></value></Parameter>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.UploadLogFile import UploadLogFile]]></body>
<origin></origin>
<description><![CDATA[Uploads the output log files]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>UploadLogFile</type>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[False]]></value></Parameter>
</ModuleDefinition>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.ComputeOutputDataList import ComputeOutputDataList
]]></body>
<origin></origin>
<description><![CDATA[Compute the output data list to be treated by the last finalization]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>ComputeOutputDataList</type>
</ModuleDefinition>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.DDSimAnalysis import DDSimAnalysis
]]></body>
<origin></origin>
<description><![CDATA[Module to run DDSim]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>DDSimAnalysis</type>
<Parameter name="randomSeed" type="int" linked_module="" linked_parameter="" in="False" out="False" description="Random seed for the generator"><value><![CDATA[0]]></value></Parameter>
<Parameter name="detectorModel" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Detecor model for simulation"><value><![CDATA[]]></value></Parameter>
<Parameter name="startFrom" type="int" linked_module="" linked_parameter="" in="False" out="False" description="From where DDSim starts to read the input file"><value><![CDATA[0]]></value></Parameter>
<Parameter name="debug" type="bool" linked_module="" linked_parameter="" in="False" out="False" description="debug mode"><value><![CDATA[False]]></value></Parameter>
</ModuleDefinition>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.ILDRegisterOutputData import ILDRegisterOutputData]]></body>
<origin></origin>
<description><![CDATA[Module to add in the metadata catalog the relevant info about the files]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>ILDRegisterOutputData</type>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[False]]></value></Parameter>
</ModuleDefinition>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.UploadOutputData import UploadOutputData]]></body>
<origin></origin>
<description><![CDATA[Uploads the output data]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>UploadOutputData</type>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[False]]></value></Parameter>
</ModuleDefinition>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.ReportErrors import ReportErrors]]></body>
<origin></origin>
<description><![CDATA[Reports errors at the end]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>ReportErrors</type>
</ModuleDefinition>
<ModuleDefinition>
<body><![CDATA[from ILCDIRAC.Workflow.Modules.FailoverRequest import FailoverRequest]]></body>
<origin></origin>
<description><![CDATA[Sends any failover requests]]></description>
<descr_short></descr_short>
<required></required>
<version>0.0</version>
<type>FailoverRequest</type>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[False]]></value></Parameter>
</ModuleDefinition>
<StepDefinition>
<origin></origin>
<version>0.0</version>
<type>ddsim_step_1</type>
<description><![CDATA[]]></description>
<descr_short></descr_short>
<Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Application Name"><value><![CDATA[]]></value></Parameter>
<Parameter name="applicationVersion" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Application Version"><value><![CDATA[]]></value></Parameter>
<Parameter name="SteeringFile" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Steering File"><value><![CDATA[]]></value></Parameter>
<Parameter name="applicationLog" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Log File"><value><![CDATA[]]></value></Parameter>
<Parameter name="ExtraCLIArguments" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Extra CLI arguments"><value><![CDATA[]]></value></Parameter>
<Parameter name="InputFile" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Input File"><value><![CDATA[]]></value></Parameter>
<Parameter name="OutputFile" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Output File"><value><![CDATA[]]></value></Parameter>
<Parameter name="OutputPath" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output File path on the grid"><value><![CDATA[]]></value></Parameter>
<Parameter name="outputPathREC" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output REC File path on the grid"><value><![CDATA[]]></value></Parameter>
<Parameter name="outputPathDST" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output DST File path on the grid"><value><![CDATA[]]></value></Parameter>
<Parameter name="OutputSE" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output File storage element"><value><![CDATA[]]></value></Parameter>
<Parameter name="listoutput" type="list" linked_module="" linked_parameter="" in="False" out="False" description="list of output file name"><value><![CDATA[[]]]></value></Parameter>
<Parameter name="ForgetInput" type="boolean" linked_module="" linked_parameter="" in="False" out="False" description="Do not overwrite input steering"><value><![CDATA[False]]></value></Parameter>
<ModuleInstance>
<type>DDSimAnalysis</type>
<name>ddsim_step_1</name>
<descr_short></descr_short>
<Parameter name="randomSeed" type="int" linked_module="" linked_parameter="" in="False" out="False" description="Random seed for the generator"><value><![CDATA[-1]]></value></Parameter>
<Parameter name="detectorModel" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Detecor model for simulation"><value><![CDATA[ILD_s5_v02]]></value></Parameter>
<Parameter name="startFrom" type="int" linked_module="" linked_parameter="" in="False" out="False" description="From where DDSim starts to read the input file"><value><![CDATA[0]]></value></Parameter>
<Parameter name="debug" type="bool" linked_module="" linked_parameter="" in="False" out="False" description="debug mode"><value><![CDATA[False]]></value></Parameter>
</ModuleInstance>
<ModuleInstance>
<type>ComputeOutputDataList</type>
<name>ddsim_step_1</name>
<descr_short></descr_short>
</ModuleInstance>
</StepDefinition>
<StepDefinition>
<origin></origin>
<version>0.0</version>
<type>Job_Finalization</type>
<description><![CDATA[]]></description>
<descr_short></descr_short>
<ModuleInstance>
<type>UploadOutputData</type>
<name>dataUpload</name>
<descr_short></descr_short>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[True]]></value></Parameter>
</ModuleInstance>
<ModuleInstance>
<type>ILDRegisterOutputData</type>
<name>ILDRegisterOutputData</name>
<descr_short></descr_short>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[True]]></value></Parameter>
</ModuleInstance>
<ModuleInstance>
<type>UploadLogFile</type>
<name>logUpload</name>
<descr_short></descr_short>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[True]]></value></Parameter>
</ModuleInstance>
<ModuleInstance>
<type>FailoverRequest</type>
<name>failoverRequest</name>
<descr_short></descr_short>
<Parameter name="enable" type="bool" linked_module="" linked_parameter="" in="True" out="False" description="EnableFlag"><value><![CDATA[True]]></value></Parameter>
</ModuleInstance>
<ModuleInstance>
<type>ReportErrors</type>
<name>reportErrors</name>
<descr_short></descr_short>
</ModuleInstance>
</StepDefinition>
<StepInstance>
<type>ddsim_step_1</type>
<name>ddsim_step_1</name>
<descr_short></descr_short>
<Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Application Name"><value><![CDATA[ddsim]]></value></Parameter>
<Parameter name="applicationVersion" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Application Version"><value><![CDATA[ILCSoft-02-00-02_gcc49]]></value></Parameter>
<Parameter name="SteeringFile" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Steering File"><value><![CDATA[ddsim_steer.py]]></value></Parameter>
<Parameter name="applicationLog" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Log File"><value><![CDATA[ddsim_ILCSoft-02-00-02_gcc49_@{STEP_ID}.log]]></value></Parameter>
<Parameter name="ExtraCLIArguments" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Extra CLI arguments"><value><![CDATA[%%20--vertexSigma%%200.0%%200.0%%200.1468%%200.0%%20--vertexOffset%%200.0%%200.0%%200.0%%200.0%%20]]></value></Parameter>
<Parameter name="InputFile" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Input File"><value><![CDATA[]]></value></Parameter>
<Parameter name="OutputFile" type="string" linked_module="" linked_parameter="" in="False" out="False" description="Output File"><value><![CDATA[%(OUTPUT_FILE)s]]></value></Parameter>
<Parameter name="OutputPath" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output File path on the grid"><value><![CDATA[/ilc/prod/ilc/mc-opt-3/ild/sim/1000-B1b_ws/6f_WWS/ILD_s5_v02/v02-00-02/]]></value></Parameter>
<Parameter name="outputPathREC" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output REC File path on the grid"><value><![CDATA[]]></value></Parameter>
<Parameter name="outputPathDST" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output DST File path on the grid"><value><![CDATA[]]></value></Parameter>
<Parameter name="OutputSE" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Output File storage element"><value><![CDATA[DESY-SRM]]></value></Parameter>
<Parameter name="listoutput" type="list" linked_module="" linked_parameter="" in="False" out="False" description="list of output file name"><value><![CDATA[[{'outputPath': '@{OutputPath}', 'outputDataSE': '@{OutputSE}', 'outputFile': '@{OutputFile}'}]]]></value></Parameter>
<Parameter name="ForgetInput" type="boolean" linked_module="" linked_parameter="" in="False" out="False" description="Do not overwrite input steering"><value><![CDATA[False]]></value></Parameter>
</StepInstance>
<StepInstance>
<type>Job_Finalization</type>
<name>finalization</name>
<descr_short></descr_short>
</StepInstance>
</Workflow>
"""  # noqa: E501


def nameTests(testcase_func, param_num, param):
  """Construct the name for the tests from the parameter tuple."""
  return '%s_%s_%s_%s' % (testcase_func.__name__, param[0][1], param[0][0].__class__.__name__, param_num)


@parameterized.expand(product(('/vo/input/file1', ['/vo/input/file1'], None),
                              ('MCSimulation', 'MCSimulation_ILD'),
                              ), name_func=nameTests)
def test_execute(inputData, jobType):
  """Test: OutputDataPolicyExecute"""
  outputFile = 'sv02-00-02.mILD_s5_v02.E1000-B1b_ws_sim.slcio'
  from ILCDIRAC.Core.Utilities.OutputDataPolicy import OutputDataPolicy
  paramDict = dict()
  paramDict['Job'] = WORKFLOW_XML.strip() % {'JOB_TYPE': jobType, 'OUTPUT_FILE': outputFile}
  paramDict['TransformationID'] = 12345
  paramDict['TaskID'] = 555
  paramDict['InputData'] = inputData
  odp = OutputDataPolicy(paramDict)
  result = odp.execute()
  assert result['OK']
  assert len(result['Value']['ProductionOutputData']) == 1
  assert result['Value']['ProductionOutputData'][0].endswith('12345_555.slcio')



"""Create text file with final calibration constants."""
from __future__ import absolute_import
import sys

ECALTOEM = sys.argv[1]
HCALTOEM = sys.argv[2]
ECALTOHAD = sys.argv[3]
HCALTOHAD = sys.argv[4]
CALIBR_ECAL_BARREL_INPUT = sys.argv[5]
CALIBR_ECAL_BARREL_INPUT2 = float(CALIBR_ECAL_BARREL_INPUT)
CALIBR_ECAL_BARREL = str(CALIBR_ECAL_BARREL_INPUT) + ' ' + str(CALIBR_ECAL_BARREL_INPUT2)
CALIBR_ECAL_ENDCAP = sys.argv[6]
CALIBR_ECAL_ENDCAP_CORR = sys.argv[7]
CALIBR_HCAL_BARREL = sys.argv[8]
CALIBR_HCAL_ENDCAP = sys.argv[9]
CALIBR_HCAL_OTHER = sys.argv[10]
MHHHE = sys.argv[11]
ECalGeVToMIP = sys.argv[12]
HCalGeVToMIP = sys.argv[13]
MuonGeVToMIP = sys.argv[14]
HCALBarrelTimeWindowMax = sys.argv[15]
HCALEndcapTimeWindowMax = sys.argv[16]
ECALBarrelTimeWindowMax = sys.argv[17]
ECALEndcapTimeWindowMax = sys.argv[18]
Output_Path = sys.argv[19]

jobList = ''

jobList += 'CalibrECALBarrel was found to be:                         '
jobList += str(CALIBR_ECAL_BARREL) + '\n'

jobList += 'CalibrECALEndcap was found to be:                         '
jobList += str(CALIBR_ECAL_ENDCAP) + '\n'

jobList += 'CalibrECALEndcapCorr was found to be:                         '
jobList += str(CALIBR_ECAL_ENDCAP_CORR) + '\n'

jobList += 'CalibrHCALBarrel was found to be:                   '
jobList += str(CALIBR_HCAL_BARREL) + '\n'

jobList += 'CalibrHCALEndcap was found to be:                   '
jobList += str(CALIBR_HCAL_ENDCAP) + '\n'

jobList += 'CalibrHCALOther was found to be:                    '
jobList += str(CALIBR_HCAL_OTHER) + '\n'

jobList += 'ECalGeVToMIP was found to be:                       '
jobList += str(ECalGeVToMIP) + '\n'

jobList += 'HCalGeVToMIP was found to be:                       '
jobList += str(HCalGeVToMIP) + '\n'

jobList += 'MuonGeVToMIP was found to be:                       '
jobList += str(MuonGeVToMIP) + '\n'

jobList += 'MaxHCalHitHadronicEnergy was found to be:           '
jobList += MHHHE + '\n'

jobList += 'ECalToEMGeVCalibration was found to be:             '
jobList += str(ECALTOEM) + '\n'

jobList += 'HCalToEMGeVCalibration was found to be:             '
jobList += str(HCALTOEM) + '\n'

jobList += 'ECalToHadGeVCalibrationBarrel was found to be:      '
jobList += str(ECALTOHAD) + '\n'

jobList += 'HCalToHadGeVCalibration was found to be:            '
jobList += str(HCALTOHAD) + '\n'

jobList += 'HCALBarrelTimeWindowMax is:                         '
jobList += HCALBarrelTimeWindowMax + '\n'

jobList += 'HCALEndcapTimeWindowMax is:                         '
jobList += HCALEndcapTimeWindowMax + '\n'

jobList += 'ECALBarrelTimeWindowMax is:                         '
jobList += ECALBarrelTimeWindowMax + '\n'

jobList += 'ECALEndcapTimeWindowMax is:                         '
jobList += ECALEndcapTimeWindowMax + '\n'

file = open(Output_Path + 'Final_Calibration.txt', 'w')
file.write(jobList)
file.close()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
SLCIOConcatenate : Helper to concatenate SLCIO files
"""

from __future__ import absolute_import
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC import S_OK, S_ERROR, gLogger

from ILCDIRAC.Interfaces.API.NewInterface.LCUtilityApplication import LCUtilityApplication
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class SLCIOConcatenate(LCUtilityApplication):
  """Helper to concatenate slcio files.

  Example:

  >>> slcioconcat = SLCIOConcatenate()
  >>> slcioconcat.setInputFile( ["slcioFile_1.slcio" , "slcioFile_2.slcio" , "slcioFile_3.slcio"] )
  >>> slcioconcat.setOutputFile("myNewSLCIOFile.slcio")
  """

  def __init__(self, paramdict=None):

    super(SLCIOConcatenate, self).__init__(paramdict)
    if not self.version:
      self.version = 'HEAD'
    self._modulename = "LCIOConcatenate"
    self.appname = 'lcio'
    self._moduledescription = 'Helper call to concatenate SLCIO files'

  def _applicationModule(self):
    m1 = self._createModuleDefinition()
    m1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    return m1

  def _applicationModuleValues(self, moduleinstance):
    moduleinstance.setValue('debug', self.debug)

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):
    """Checks that all needed parameters are set."""

    if not self.outputFile and self._jobtype == 'User':
      self.setOutputFile('LCIOFileConcatenated.slcio')
      LOG.notice('No output file name specified. Output file : LCIOFileConcatenated.slcio')

    if self._jobtype != 'User':
      self._listofoutput.append({"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})

    #res = self._checkRequiredApp()
    # if not res['OK']:
    #  return res

    return S_OK()

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for the CalibrationSystem/Utilities/functions.py."""

from __future__ import print_function
from __future__ import absolute_import
import pytest
import os
import string
import random
import tempfile
import shutil

from ILCDIRAC.CalibrationSystem.Utilities.functions import readParameterDict
from ILCDIRAC.CalibrationSystem.Utilities.functions import readParametersFromSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import updateSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import addParameterToProcessor
from six.moves import range

__RCSID__ = "$Id$"
MODULE_NAME = 'ILCDIRAC.CalibrationSystem.Utilities.functions'


def copySteeringFile(tag, calibID):
  """Copy steering files to local test directory."""
  workdirName = 'calib%s' % calibID
  if not os.path.exists(workdirName):
    os.makedirs(workdirName)

  if tag == 'CLIC':
    src = os.path.join(os.environ['ILCDIRAC_BASE_FOLDER'], 'Tests', 'Files', 'clicReconstruction.xml')
    shutil.copyfile(src, '%s/clicReconstruction.xml' % workdirName)
    return '%s/clicReconstruction.xml' % workdirName
  elif tag == 'FCCee':
    src = os.path.join(os.environ['ILCDIRAC_BASE_FOLDER'], 'Tests', 'Files', 'fccReconstruction.xml')
    shutil.copyfile(src, '%s/fccReconstruction.xml' % workdirName)
    return '%s/fccReconstruction.xml' % workdirName
  else:
    return None


def cleanDir(calibID):
  """Remove test directory."""
  workdirName = 'calib%s' % calibID
  if os.path.exists(workdirName):
    try:
      shutil.rmtree(workdirName)
    except EnvironmentError as e:
      print("Failed to delete directory: %s; ErrMsg: %s" % (workdirName, str(e)))
      assert False


@pytest.fixture
def copyFccSteeringFile():
  """Copy FCC steering file."""
  calibID = 1
  yield copySteeringFile('FCCee', calibID)
  #  cleanDir(calibID)


@pytest.fixture
def copyClicSteeringFile():
  """Copy CLIC steering file."""
  calibID = 1
  yield copySteeringFile('CLIC', calibID)
  cleanDir(calibID)


@pytest.fixture
def produceRandomTextFile():
  """Produce random text."""
  f = tempfile.NamedTemporaryFile(delete=False)
  nLines = random.randint(2, 20)
  for _ in range(0, nLines):
    nSymbolsInLine = random.randint(0, 120)
    line = ''
    for _ in range(0, nSymbolsInLine):
      line += random.choice(string.ascii_letters + '       ')
    f.write(line.encode())
  f.close()
  yield f.name
  os.unlink(f.name)


@pytest.fixture
def readEmptyParameterDict():
  """Read parameters from the file."""
  import ILCDIRAC.CalibrationSystem.Utilities as utilities
  fileDir = os.path.join(utilities.__path__[0], 'auxiliaryFiles')

  inFileName = os.path.join(fileDir, 'parameterListMarlinSteeringFile.txt')
  parDict = readParameterDict(inFileName)
  for iKey in list(parDict.keys()):
    if 'RootFile' in iKey:
      parDict.pop(iKey, None)
  return parDict


def test_addParameterToProcessor(produceRandomTextFile, copyFccSteeringFile, mocker):
  """Test adding of parameter to processor in Marlin steering file."""
  # non-existing input file
  res = addParameterToProcessor('dummy.xml', 'dummyProc', {'name': 'dummyValue'})
  assert not res['OK']
  assert "cannot find input" in res['Message']
  # non-xml input file
  randomFile = produceRandomTextFile
  res = addParameterToProcessor(randomFile, 'dummyProc', {'name': 'dummyValue'})
  assert not res['OK']
  assert "cannot parse input" in res['Message']
  # good input file, non-existing processor
  steeringFile = copyFccSteeringFile
  res = addParameterToProcessor(steeringFile, 'dummyProc', {'name': 'dummyValue'})
  assert not res['OK']
  assert "Can't find processor" in res['Message']
  # good input file, good processor name, no 'name' key in the parameter dict
  steeringFile = copyFccSteeringFile
  res = addParameterToProcessor(steeringFile, 'dummyProc', {'dummy': 'dummyValue'})
  assert not res['OK']
  assert "parameter dict should have key 'name'" in res['Message']
  # good input file, good processor name
  res = addParameterToProcessor(steeringFile, 'MyAIDAProcessor', {'name': 'dummyValue'})
  assert res['OK']
  # good input file, good processor name, second append of the parameter with the same name
  res = addParameterToProcessor(steeringFile, 'MyAIDAProcessor', {'name': 'dummyValue'})
  assert not res['OK']
  assert ("parameter with name %s already exists" % 'dummyValue') in res['Message']
  # good input file, good processor name
  res = addParameterToProcessor(steeringFile, 'MyDDCaloDigi_10ns', {'name': 'ECALLayers', 'type': 'IntVec',
                                                                    'value': '10 31'})
  assert res['OK']


def test_updateSteeringFile(copyClicSteeringFile, readEmptyParameterDict):
  """Test updateSteeringFile."""
  initialParDict = readEmptyParameterDict

  parDict1 = dict(initialParDict)
  #  inFileName = os.path.join(self.fileDir, 'clicReconstruction_2019-04-17.xml')
  inFileName = copyClicSteeringFile
  res = readParametersFromSteeringFile(inFileName, parDict1)
  #  key1 = "processor[@name='MyPfoAnalysis']/parameter[@name='RootFile']"
  #  parDict1[key1] = "dummyDummyRootFile.root"
  #  key2 = "global/parameter[@name='LCIOInputFiles']"
  #  parDict1[key2] = "in1.slcio, in2.slcio"
  #  self.assertTrue(len(parDict1) == len(initialParDict),
  #                  "two dictionaries have to be the same size. len1: %s; len2: %s"
  #                  % (len(parDict1), len(initialParDict)))

  outFileName = os.path.join(os.path.dirname(inFileName), 'out1.xml')
  res = updateSteeringFile(inFileName, outFileName, parDict1)
  assert res['OK']

  parDict2 = dict(initialParDict)
  res = readParametersFromSteeringFile(outFileName, parDict2)
  assert len(parDict1) == len(parDict2)

  notEqualValues = False
  for iKey in initialParDict:
    if parDict1[iKey] != parDict2[iKey]:
      notEqualValues = True
  assert not notEqualValues


def test_readParameterDict(readEmptyParameterDict):
  """Test readParameterDict."""
  parDict = readEmptyParameterDict
  assert '' not in parDict

  allValuesAreNone = True
  for _, iVal in parDict.items():
    if iVal is not None:
      allValuesAreNone = False
  assert allValuesAreNone


def test_readParametersFromSteeringFile(copyClicSteeringFile, readEmptyParameterDict):
  """Test readParametersFromSteeringFile."""
  parDict = readEmptyParameterDict
  inFileName = copyClicSteeringFile
  res = readParametersFromSteeringFile(inFileName, parDict)
  print(res)
  assert res['OK']

  someValuesAreNone = False
  for _, iVal in parDict.items():
    if iVal is None:
      someValuesAreNone = True
  assert not someValuesAreNone


def test_splitFilesAcrossJobs(mocker):
  """Test splitting of file across jobs."""
  from ILCDIRAC.CalibrationSystem.Utilities.functions import splitFilesAcrossJobs
  inputFiles = {'muon': ['muon1', 'muon2', 'muon3', 'muon4', 'muon5'],
                'kaon': ['kaon1', 'kaon2', 'kaon3', 'kaon4', 'kaon5'],
                'gamma': ['gamma1', 'gamma2', 'gamma3', 'gamma4', 'gamma5'],
                'zuds': ['zuds1', 'zuds2', 'zuds3', 'zuds4', 'zuds5']}
  nEventsPerFile = {'muon': 20, 'kaon': 24, 'gamma': 25, 'zuds': 30}

  def printOut(nJobs):
    print("\nnEventsPerFile: %s" % nEventsPerFile)
    print("nTotalEvents:")
    for iKey, iFiles in inputFiles.items():
      print("%s: %s" % (iKey, len(iFiles) * nEventsPerFile[iKey]))
    print("")

    outDict = splitFilesAcrossJobs(inputFiles, nEventsPerFile, nJobs)
    for i in range(0, nJobs):
      print("Job #%s:" % i)
      for iKey, iVal in outDict[i].items():
        print("%s\t --> %s" % (iKey, iVal))

  nJobs = 5
  printOut(nJobs)
  outDict = splitFilesAcrossJobs(inputFiles, nEventsPerFile, nJobs)
  for i in range(0, nJobs):
    for iKey, iVal in outDict[i].items():
      assert len(iVal[0]) == 1
      assert iVal[1] == 0
      assert iVal[2] == nEventsPerFile[iKey]

  nJobs = 2
  printOut(nJobs)
  outDict = splitFilesAcrossJobs(inputFiles, nEventsPerFile, nJobs)
  for i in range(0, nJobs):
    for iKey, iVal in outDict[i].items():
      assert len(iVal[0]) == 3
      assert iVal[1] == 0 or iVal[1] == nEventsPerFile[iKey] // 2
      assert iVal[2] == len(inputFiles[iKey]) * nEventsPerFile[iKey] // 2

  #  assert False



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
SLICPandora : Run Pandora in the SID context
"""

from __future__ import absolute_import
import os
import types

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.InstalledFiles import Exists
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class SLICPandora(LCApplication):
  """Call SLICPandora.

  Usage:

  >>> lcsim = LCSIM()
  ...
  >>> slicpandora = SLICPandora()
  >>> slicpandora.getInputFromApp(lcsim)
  >>> slicpandora.setPandoraSettings("~/GreatPathToHeaven/MyPandoraSettings.xml")
  >>> slicpandora.setStartFrom(10)

  Use :func:`setExtraCLIArguments` if you want to add arguments to the PandoraFrontend call
  """

  def __init__(self, paramdict=None):

    self.startFrom = 0
    self.pandoraSettings = ''
    self.detectorModel = ''
    super(SLICPandora, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'SLICPandoraAnalysis'
    self._moduledescription = 'Module to run SLICPANDORA'
    self.appname = 'slicpandora'
    self.datatype = 'REC'
    self.detectortype = 'SID'
    self._paramsToExclude.extend(["outputDstPath", "outputRecPath", "OutputDstFile", "OutputRecFile"])

  def setDetectorModel(self, detectorModel):
    """Define detector to use for SlicPandora simulation.

    :param str detectorModel: Detector Model to use for SlicPandora simulation.
    """
    self._checkArgs({'detectorModel': (str,)})

    self.detectorModel = detectorModel
    if os.path.exists(detectorModel) or detectorModel.lower().count("lfn:"):
      self.inputSB.append(detectorModel)

  def setStartFrom(self, startfrom):
    """ Optional: Define from where slicpandora start to read in the input file

    :param int startfrom: from how slicpandora start to read the input file
    """
    self._checkArgs({'startfrom': int})
    self.startFrom = startfrom

  def setPandoraSettings(self, pandoraSettings):
    """ Optional: Define the path where pandora settings are

    :param str pandoraSettings: path where pandora settings are
    """
    self._checkArgs({'pandoraSettings': (str,)})
    self.pandoraSettings = pandoraSettings
    if os.path.exists(pandoraSettings) or pandoraSettings.lower().count("lfn:"):
      self.inputSB.append(pandoraSettings)

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):

    if not self.version:
      return S_ERROR('No version found')

    if self.steeringFile:
      if not os.path.exists(self.steeringFile) and not self.steeringFile.lower().count("lfn:"):
        res = Exists(self.steeringFile)
        if not res['OK']:
          return res

    if not self.pandoraSettings:
      return S_ERROR("PandoraSettings not set, you need it")

    #res = self._checkRequiredApp()
    # if not res['OK']:
    #  return res

    if not self.startFrom:
      LOG.info('No startFrom defined for SlicPandora : start from the begining')

    if self._jobtype != 'User':
      self.prodparameters['slicpandora_steeringfile'] = self.steeringFile
      self.prodparameters['slicpandora_detectorModel'] = self.detectorModel

    return S_OK()

  def _applicationModule(self):

    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("pandorasettings", "", "string", "", "", False, False,
                               "Pandora Settings"))
    md1.addParameter(Parameter("detectorxml", "", "string", "", "", False, False,
                               "Detector model for simulation"))
    md1.addParameter(Parameter("startFrom", 0, "int", "", "", False, False,
                               "From how SlicPandora start to read the input file"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False,
                               "debug mode"))
    return md1

  def _applicationModuleValues(self, moduleinstance):

    moduleinstance.setValue("pandorasettings", self.pandoraSettings)
    moduleinstance.setValue("detectorxml", self.detectorModel)
    moduleinstance.setValue("startFrom", self.startFrom)
    moduleinstance.setValue("debug", self.debug)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run DDSim.

:author: Andre Sailer
:since:  November 16, 2015
"""

from __future__ import absolute_import
import os

from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC import S_OK, S_ERROR, gLogger
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getEnvironmentScript, getSoftwareFolder
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import getNewLDLibs
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import resolveIFpaths, getProdFilename
from ILCDIRAC.Workflow.Utilities.DD4hepMixin import DD4hepMixin
from ILCDIRAC.Core.Utilities.FindSteeringFileDir import getSteeringFileDirName
import six


__RCSID__ = '$Id$'
DDSIMINPUTFORMATS = ('.stdhep', '.hepevt', '.HEPEvt', '.slcio', '.hepmc')
LOG = gLogger.getSubLogger(__name__)


class DDSimAnalysis(DD4hepMixin, ModuleBase):
  """Specific Module to run a DDSim job."""

  def __init__(self):
    super(DDSimAnalysis, self).__init__()
    self.enable = True
    self.STEP_NUMBER = ''
    self.result = S_ERROR()
    self.applicationName = 'ddsim'
    self.startFrom = 0
    self.randomSeed = -1
    self.detectorModel = ''
    self.extraParticles = []
    self.eventstring = ['+++ Initializing event']

  def applicationSpecificInputs(self):
    """Resolve all input variables for the module here.

    :return: S_OK()
    """

    if self.WorkflowStartFrom:
      self.startFrom = self.WorkflowStartFrom

    self.randomSeed = self._determineRandomSeed()

    if "IS_PROD" in self.workflow_commons and self.workflow_commons["IS_PROD"]:
      self.OutputFile = getProdFilename(self.OutputFile,
                                        int(self.workflow_commons["PRODUCTION_ID"]),
                                        int(self.workflow_commons["JOB_ID"]),
                                        self.workflow_commons,
                                       )

    if not len(self.InputFile) and len(self.InputData):
      for files in self.InputData:
        if files.endswith(DDSIMINPUTFORMATS):
          self.InputFile.append(files)

    return S_OK('Parameters resolved')

  def resolveDDSimOptions(self, envScriptPath, compactFile):
    """Find all the options combining CLI arguments and steering file.

    :return: S_OK(list of strings of the dump of the ddsim steering file)
    """
    scriptName = 'findOptions.sh'
    script = []
    script.append('#!/bin/bash')
    script.append('############################################################')
    script.append('# Dynamically generated script to resolve DDSim parameters #')
    script.append('############################################################')
    script.append('source %s' % envScriptPath)
    script.append('echo =========')
    com1 = 'ddsim --dumpSteeringFile'
    com2 = '--compactFile %(compact)s %(extraArgs)s >& ddsimDump.txt' % dict(compact=compactFile,
                                                                             extraArgs=self.extraCLIarguments)
    script.append('%s %s' % (com1, com2))
    script.append('declare -x appstatus=$?')
    script.append('exit $appstatus')

    with open(scriptName, 'w') as scriptFile:
      scriptFile.write("\n".join(script))

    os.chmod(scriptName, 0o755)
    comm = 'bash "./%s"' % scriptName
    result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    if not result['OK']:
      return S_ERROR("Could not resolve DDSim options, shell call failed.")

    options = []
    with open("ddsimDump.txt", "r") as ddsimOptions:
      options = ddsimOptions.read().splitlines()
    if not options:
      return S_ERROR("DDSim steering file dump empty")
    return S_OK(options)

  def createParticleTable(self, envScriptPath, options):
    """Resolve the particle.tbl and add or modify the particles.

    :return: S_OK(list of strings of the new/modified particle table)
    """
    # Check if in the options a particle.tbl is set
    particleTableLocation = None
    for line in options:
      if "SIM.physics.pdgfile =" in line:
        if "SIM.physics.pdgfile = None" in line:
          particleTableLocation = None
        else:
          particleTableLocation = line.split('=')[1].strip('" ')

    if not particleTableLocation:
      # Get the env file and extract the location of DD4hep
      dd4hepLocation = None
      with open(envScriptPath, "r") as envScrip:
        for line in envScrip:
          if "DD4hepINSTALL" in line:
            dd4hepLocation = line.split('\"', 1)[-1].split('\"', 1)[0]

        if dd4hepLocation is None:
          return S_ERROR("Did not find env variable DD4hepINSTALL")
        # Get the particle table
        particleTableLocation = "%s/DDG4/examples/particle.tbl" % dd4hepLocation

    if not os.path.exists(particleTableLocation):
      return S_ERROR("Cannot find particle table at location %s" % particleTableLocation)

    particleTable = None
    with open(particleTableLocation, "r") as particleTableFile:
      particleTable = particleTableFile.readlines()

    # First we remove the particles from the table that we want to modify
    newParticleTable = []
    removeIDs = {str(extraParticle[0]) for extraParticle in self.extraParticles}
    for particle in particleTable:
      if particle.split()[0] not in removeIDs:
        newParticleTable.append(particle.rstrip())

    # Add new particles to table
    for particle in self.extraParticles:
      newParticleTable.append(' '.join([str(f) for f in particle]))
    return S_OK(newParticleTable)

  def runIt(self):
    """Called by JobAgent.

    Execute the following:
      - get the environment variables that should have been set during installation
      - find the detector model xml, using CS query to obtain the path
      - prepare the steering file and command line parameters
      - run DDSim on this steering file and catch the exit status

    :returns: :func:`~DIRAC.Core.Utilities.ReturnValues.S_OK`, :func:`~DIRAC.Core.Utilities.ReturnValues.S_ERROR`
    """
    self.result = S_OK()
    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')
    if not self.result['OK']:
      LOG.error("Failed to resolve input parameters:", self.result['Message'])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('DDSim should not proceed as previous step did not end properly')

    # TODO: Setup LD_LIBRARY_PATH for extensions
    res = getEnvironmentScript(self.platform, self.applicationName, self.applicationVersion, self.getEnvScript)
    if not res['OK']:
      LOG.error("Could not obtain the environment script: ", res["Message"])
      return res
    envScriptPath = res["Value"]

    # get the path to the detector model, either local or from the software
    resXML = self._getDetectorXML()
    if not resXML['OK']:
      LOG.error("Could not obtain the detector XML file: ", resXML["Message"])
      return resXML
    compactFile = resXML['Value']

    if len(self.InputFile):
      res = resolveIFpaths(self.InputFile)
      if not res['OK']:
        LOG.error("Generator file not found")
        return res
      self.InputFile = res['Value']

    # if steering file is set try to find it
    if len(self.SteeringFile) > 0:
      self.SteeringFile = os.path.basename(self.SteeringFile)
      if not os.path.exists(self.SteeringFile):
        res = getSteeringFileDirName(self.platform, self.applicationName, self.applicationVersion)
        if not res['OK']:
          LOG.error("Could not find where the steering files are")
          return res
        steeringfiledirname = res['Value']
        if os.path.exists(os.path.join(steeringfiledirname, self.SteeringFile)):
          self.SteeringFile = os.path.join(steeringfiledirname, self.SteeringFile)
      if not os.path.exists(self.SteeringFile):
        LOG.error("Missing steering file")
        return S_ERROR("Could not find steering file")
      self.extraCLIarguments += " --steeringFile %s " % self.SteeringFile

    if self.startFrom:
      self.extraCLIarguments += " --skipNEvents %s " % self.startFrom

    if self.debug:
      self.extraCLIarguments += " --printLevel DEBUG "

    # Same as for mokka: using ParticleGun does not imply InputFile
    if self.InputFile:
      self.InputFile = [self.InputFile] if isinstance(self.InputFile, six.string_types) else self.InputFile
      self.extraCLIarguments += " --inputFile %s " % self.InputFile[0]

    if self.NumberOfEvents:
      self.extraCLIarguments += " --numberOfEvents %s " % self.NumberOfEvents

    self.extraCLIarguments += " --random.seed %s " % self.randomSeed

    if self.OutputFile:
      self.extraCLIarguments += " --outputFile %s " % self.OutputFile

    if self.extraParticles:
      res = self.resolveDDSimOptions(envScriptPath, compactFile)
      if not res['OK']:
        return res
      resTbl = self.createParticleTable(envScriptPath, res['Value'])
      if not resTbl['OK']:
        return resTbl
      with open("particle.tbl", 'w') as tableFile:
        tableFile.write("\n".join(resTbl['Value']))
      self.extraCLIarguments = "%s --physics.pdgfile particle.tbl" % self.extraCLIarguments

    scriptName = 'DDSim_%s_Run_%s.sh' % (self.applicationVersion, self.STEP_NUMBER)
    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = []
    script.append('#!/bin/bash')
    script.append('#####################################################################')
    script.append('# Dynamically generated script to run a production or analysis job. #')
    script.append('#####################################################################')
    script.append('source %s' % envScriptPath)
    script.append('echo =========')

    # for user provided libraries
    if os.path.exists("lib"):
      script.append("export LD_LIBRARY_PATH=$PWD/lib:$LD_LIBRARY_PATH")

    script.append('env | sort >> localEnv.log')
    script.append('echo ddsim:`which ddsim`')
    script.append('echo =========')
    comm = 'ddsim --compactFile %(compactFile)s %(extraArgs)s' % dict(compactFile=compactFile,
                                                                      extraArgs=self.extraCLIarguments)
    LOG.info("Command:", comm)
    script.append(comm)
    script.append('declare -x appstatus=$?')
    script.append('exit $appstatus')

    with open(scriptName, 'w') as scriptFile:
      scriptFile.write("\n".join(script))

    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    os.chmod(scriptName, 0o755)
    comm = 'bash "./%s"' % scriptName
    self.setApplicationStatus('DDSim %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    resultTuple = self.result['Value']
    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed to produce log file' % (self.applicationName))
      if not self.ignoreapperrors:
        return S_ERROR('%s did not produce the expected log %s' % (self.applicationName, self.applicationLog))
    status = resultTuple[0]

    LOG.info("Status after the application execution is %s" % status)

    return self.finalStatusReport(status)

  def getEnvScript(self, platform, appname, appversion):
    """Create an environment script for ddsim. Only used when CVMFS native installation not available.

    We need ROOTSYS, G4INSTALL and G4DATA as additional environment variables in the CS
    We need to set DD4hepINSTALL !
    """
    # Need to fetch the new LD_LIBRARY_PATH
    newLDLibraryPath = getNewLDLibs(platform, appname, appversion)
    softwareFolder = getSoftwareFolder(platform, appname, appversion)
    if not softwareFolder['OK']:
      return softwareFolder
    softwareRoot = softwareFolder['Value']
    envName = "DDSimEnv.sh"

    script = []
    script.append("#!/bin/bash")
    script.append("##########################")
    script.append("## Env script for DDSim ##")
    script.append("##########################")

    addEnv = self.ops.getOptionsDict("/AvailableTarBalls/%s/%s/%s/AdditionalEnvVar" % (platform,
                                                                                       appname,
                                                                                       appversion))

    if addEnv['OK']:
      for variable, value in addEnv['Value'].items():
        script.append('declare -x %s=%s' % (variable, value))
    else:
      LOG.verbose("No additional environment variables needed for this application")

    # if these variables are not given in the configuration system we guess them from the G4DATA folder
    for var, folder in {"G4LEDATA": "G4EMLOW",
                         "G4LEVELGAMMADATA": "PhotonEvaporation",
                         "G4NEUTRONXSDATA": "G4NEUTRONXS",
                         "G4SAIDXSDATA": "G4SAIDDATA",
                         "G4RADIOACTIVEDATA": "RadioactiveDecay",
                         "G4NEUTRONHPDATA": "G4NDL",
                       }.items():
      if var not in addEnv.get('Value', {}):
        script.append('declare -x %s=$(ls -d $G4DATA/%s*)' % (var, folder))

    init = self.ops.getValue("/AvailableTarBalls/%s/%s/%s/InitScript" % (platform,
                                                                         appname,
                                                                         appversion), None)
    if init:
      script.append('source %s' % init)

    # Executable:
    script.append('declare -x PATH=%s/bin:$PATH' % softwareRoot)
    script.append('declare -x DD4hepINSTALL=%s' % softwareRoot)

    # Python objects, pyroot
    script.append('declare -x PYTHONPATH=%s/lib/python:$PYTHONPATH' % softwareRoot)
    script.append('declare -x PYTHONPATH=$ROOTSYS/lib:$PYTHONPATH')

    # For ROOT CLING Stuff
    script.append('declare -x CPATH=%s/include:$CPATH' % softwareRoot)

    # Libraries
    if newLDLibraryPath:
      script.append('declare -x LD_LIBRARY_PATH=$LD_LIBRARY_PATH:%s' % newLDLibraryPath)

    # user provided libraries are in lib in the job working directory
    if os.path.exists("%s/lib" % os.getcwd()):
      script.append('declare -x LD_LIBRARY_PATH=%s/lib:$LD_LIBRARY_PATH' % os.getcwd())

    # Root Path, just in case
    script.append('declare -x PATH=$ROOTSYS/bin:$PATH')

    # Root and Geant4 Library Path
    script.append('declare -x LD_LIBRARY_PATH=$ROOTSYS/lib:$LD_LIBRARY_PATH')
    script.append('declare -x LD_LIBRARY_PATH=$G4INSTALL/lib64:$LD_LIBRARY_PATH')
    script.append('declare -x LD_LIBRARY_PATH=$DD4hepINSTALL/lib:$LD_LIBRARY_PATH')

    with open(envName, "w") as scriptFile:
      scriptFile.write("\n".join(script))
      scriptFile.write("\n")

    os.chmod(envName, 0o755)
    return S_OK(os.path.abspath(envName))

  def _determineRandomSeed(self):
    """determine what the randomSeed should be, depends on production or not.

    .. Note::
      DDSim we use *randomSeed* and not *RandomSeed* as in the other workflow modules
    """
    if self.randomSeed == -1:
      self.randomSeed = self.jobID
    if "IS_PROD" in self.workflow_commons:
      self.randomSeed = int(
          str(
              int(
                  self.workflow_commons["PRODUCTION_ID"]))
          + str(
              int(
                  self.workflow_commons["JOB_ID"])))
    return self.randomSeed



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""List the VO members.

Accesses the VOMS server to print members of given VO

Options:
  -u, --UserName <value>      Family name of the user
  -v, --VO <value>            VO to print or search: [ilc|calice]
  -A, --addUser               print output as input for dirac-ilc-add-user
"""

from __future__ import print_function
from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger, exit as dexit
from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

LOG = gLogger.getSubLogger(__name__)

# Check if suds exists and print information how to install it
try:
  import suds  # pylint: disable=W0611
except ImportError:
  LOG.error("Run pip install suds_jurko")
  raise


class _Params(object):
  """Parameter Object."""

  def __init__(self):
    self.username = ''
    self.voName = 'ilc'
    self.adminUrl = "https://grid-voms.desy.de:8443/voms/%s/services/VOMSAdmin"
    self.attributeUrl = "https://grid-voms.desy.de:8443/voms/%s/services/VOMSAttributes"
    self.addPrint = False

  def registerSwitches(self):
    Script.registerSwitch("u:", "UserName=", "Family name of the user", self.setUser)
    Script.registerSwitch("v:", "VO=", "VO to print or search: [ilc|calice]", self.setVO)
    Script.registerSwitch("A", "addUser:", "print output as input for dirac-ilc-add-user", self.setAddPrint)
    Script.setUsageMessage("""%s -U <username> [-v ilc|calice] [-A]""" % Script.scriptName)

  def setUser(self, opt):
    self.username = opt
    return S_OK()

  def setVO(self, opt):
    if opt not in ['ilc', 'calice']:
      return S_ERROR("Unknown VO %s: ilc or calice only" % opt)
    self.voName = opt
    return S_OK()

  def setAddPrint(self, dummy=False):
    """Set the flag to print user strings as input for dirac-ilc-add-user."""
    self.addPrint = True
    return S_OK()

  def setURLs(self):
    """Set the proper urls based on the vo."""
    self.adminUrl = self.adminUrl % self.voName
    self.attributeUrl = self.attributeUrl % self.voName


def printUser(user, addPrint):
  """print user information."""
  if addPrint:
    gLogger.notice("-D\"%s\" -C\"%s\" -E\"%s\"" % (user['DN'], user['CA'], user['mail']))
  else:
    gLogger.notice("%s, %s, %s" % (user['DN'], user['CA'], user['mail']))


@DIRACScript()
def main():
  """Print the list of users in the VO."""
  clip = _Params()
  clip.registerSwitches()
  Script.parseCommandLine()
  clip.setURLs()

  from DIRAC.Core.Security.VOMSService import VOMSService
  voms = VOMSService(vo=clip.voName)
  res = voms.getUsers()
  if not res['OK']:
    gLogger.error(res['Message'])
    dexit(1)
  users = res['Value']
  for userDN, userInfo in users.items():
    userInfo['DN'] = userDN
    if not clip.username:
      printUser(userInfo, clip.addPrint)
    else:
      if userDN.lower().count(clip.username.lower()):
        printUser(userInfo, clip.addPrint)


if __name__ == "__main__":
  main()



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test Whizard2 module."""

from __future__ import print_function
from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from parameterized import parameterized
import six

from DIRAC import gLogger, S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Whizard2
from Tests.Utilities.FileUtils import mock_open_few
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds


__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.Whizard2'
BUILTIN_NAME = 'builtins' if six.PY3 else '__builtin__'

gLogger.setLevel("DEBUG")
gLogger.showHeaders(True)

# pylint: disable=protected-access


class Whizard2TestCase(unittest.TestCase):
  """Base class for the Whizard2 test cases."""

  def setUp(self):
    """set up the objects."""
    self.whiz = Whizard2({})
    self.whiz._ops = Mock(name='OpsMock')

  def test_setrandomseed(self):
    self.assertFalse(self.whiz._errorDict)
    self.whiz.setRandomSeed(89421)
    self.assertFalse(self.whiz._errorDict)
    assertEqualsImproved(self.whiz.randomSeed, 89421, self)

  def test_setrandomseed_fails(self):
    self.assertFalse(self.whiz._errorDict)
    self.whiz.setRandomSeed(['abc'])
    self.assertIn('_checkArgs', self.whiz._errorDict)

  def test_setEvtType(self):
    self.assertFalse(self.whiz._errorDict)
    self.whiz.setEvtType('ee->ff')
    self.assertFalse(self.whiz._errorDict)
    assertEqualsImproved(self.whiz.eventType, 'ee->ff', self)

  @parameterized.expand([(15, False, '_checkArgs'),
                         ('tt', True, 'setEvtType'),
                         ])
  def test_setEvtType_fail(self, evtType, addedToJob, errorMessage):
    self.assertFalse(self.whiz._errorDict)
    self.whiz.addedtojob = addedToJob
    self.whiz.setEvtType(evtType)
    self.assertIn(errorMessage, self.whiz._errorDict)

  @parameterized.expand([(True,), (False,)])
  def test_setSinFile(self, raw):
    """Test setSinFile."""
    self.assertFalse(self.whiz._errorDict)
    with patch('os.path.isfile', new=Mock(return_value=True)), \
         patch('%s.open' % BUILTIN_NAME, mock_open_few({'/some/path': 'process decay_proc = "A", "A" => "b", "B"'})):
      if raw:
        self.whiz.setSinFile('/some/path', raw=True)
      else:
        self.whiz.setSinFile('/some/path', raw=False)
    self.assertFalse(self.whiz._errorDict)
    assertEqualsImproved(self.whiz.whizard2SinFile, 'process decay_proc = "A", "A" => "b", "B"', self)
    assertEqualsImproved(self.whiz.whizard2RawSin, raw, self)

  @parameterized.expand([(False, 'does not exist', ''),
                         (True, 'Do not set n_events', '\n n_events'),
                         (True, 'Do not set seed', '\n seed'),
                         (True, 'Do not call "simulate ()"', '\n simulate( decay_proc)'),
                         ])
  def test_setSinFile_fails2(self, exists, errorMessage, extra):
    with patch('os.path.isfile', new=Mock(return_value=exists)), \
        patch('%s.open' % BUILTIN_NAME,
              mock_open_few({'/some/path': 'process decay_proc = "A", "A" => "b", "B"' + extra})):
      assertDiracFailsWith(self.whiz.setSinFile('/some/path'), errorMessage, self)

  def test_checkworkflow_app_missing(self):
    self.whiz._inputapp = ['some_depdency', 'unavailable_dependency_fail_on_this']
    self.whiz._jobapps = ['myjobapp_1', 'some_dependency']
    assertDiracFailsWith(self.whiz._checkWorkflowConsistency(), 'job order not correct', self)

  def test_checkworkflow_empty(self):
    self.whiz._inputapp = []
    self.whiz._jobapps = []
    assertDiracSucceeds(self.whiz._checkWorkflowConsistency(), self)

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.whiz._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.whiz._prodjobmodules(module_mock), self)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.whiz._userjobmodules(None),
                            'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.whiz._prodjobmodules(None),
                            'prodjobmodules failed', self)

  def test_checkconsistency(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'mymodel.sin \n decay_proc'
    self.whiz.outputFile = 'myoutput.stdhep'
    self.whiz._jobtype = 'User'
    self.whiz.numberOfEvents = 100
    assertDiracSucceeds(self.whiz._checkConsistency(Mock()), self)
    self.assertNotIn({'outputFile': '@{OutputFile}', 'outputPath': '@{OutputPath}',
                        'outputDataSE': '@{OutputSE}'}, self.whiz._listofoutput)
    self.assertNotIn('nbevts', self.whiz.prodparameters)
    self.assertNotIn('Process', self.whiz.prodparameters)
    self.assertNotIn('Model', self.whiz.prodparameters)
    self.assertNotIn('Energy', self.whiz.prodparameters)

  def test_checkconsistency_noVersion(self):
    self.whiz.version = None
    assertDiracFailsWith(self.whiz._checkConsistency(Mock()), 'No version found!', self)

  def test_checkconsistency_noSinFile(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = None
    assertDiracFailsWith(self.whiz._checkConsistency(Mock()), 'No sin file set!', self)

  def test_checkconsistency_noNumberOfEvents(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'mymodel.sin'
    self.whiz.numberOfEvents = None
    assertDiracFailsWith(self.whiz._checkConsistency(Mock()), 'Number of events not set!', self)

  def test_checkconsistency_nouserjob(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'sqrts=350 GeV\n model=SM \ndecay_proc'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    assertDiracSucceeds(self.whiz._checkConsistency(Mock()), self)
    self.assertIn({'outputFile': '@{OutputFile}', 'outputPath': '@{OutputPath}',
                     'outputDataSE': '@{OutputSE}'}, self.whiz._listofoutput)
    for keyword in ['Process', 'Energy', 'nbevts', 'Model']:
      self.assertIn(keyword, self.whiz.prodparameters)
    assertEqualsImproved(self.whiz.prodparameters['Energy'], '350', self)
    assertEqualsImproved(self.whiz.prodparameters['Process'], 'ee -> ff', self)
    assertEqualsImproved(self.whiz.prodparameters['Model'], 'SM', self)
    assertEqualsImproved(self.whiz.prodparameters['nbevts'], 100, self)

  def test_checkconsistency_nouserjob_2(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = '  sqrts  =   350.232   GeV\n model=SM \ndecay_proc'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    self.whiz.energy = 1337
    assertDiracSucceeds(self.whiz._checkConsistency(Mock()), self)
    self.assertIn({'outputFile': '@{OutputFile}', 'outputPath': '@{OutputPath}',
                   'outputDataSE': '@{OutputSE}'}, self.whiz._listofoutput)
    for keyword in ['Process', 'Energy', 'nbevts', 'Model']:
      self.assertIn(keyword, self.whiz.prodparameters)
    assertEqualsImproved(self.whiz.prodparameters['Energy'], '1337', self)
    assertEqualsImproved(self.whiz.prodparameters['Process'], 'ee -> ff', self)
    assertEqualsImproved(self.whiz.prodparameters['Model'], 'SM', self)
    assertEqualsImproved(self.whiz.prodparameters['nbevts'], 100, self)
    self.assertIn("sqrts = 1337 GeV", self.whiz.whizard2SinFile)

  def test_checkconsistency_nouserjob_fails(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'mymodel.sin \ndecay_proc'
    self.whiz.numberOfEvents = 100
    self.whiz._jobtype = 'notUser'
    assertDiracFailsWith(self.whiz._checkConsistency(Mock()), 'evttype not set, please set event type!', self)

  def test_checkconsistency_nouserjob_fails2(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'model=SM \ndecay_proc'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    assertDiracFailsWith(
        self.whiz._checkConsistency(
            Mock()),
        'No energy set in sin file, please set "sqrts=...GeV"',
        self)

  def test_checkconsistency_nouserjob_fails3(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'sqrts=350 GeV \ndecay_proc'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    assertDiracFailsWith(self.whiz._checkConsistency(Mock()), 'No model set in sin file, please set "model=..."', self)

  def test_checkconsistency_nouserjob_fails4(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'sqrts=350 GeV\n sqrts=550 GeV \ndecay_proc'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    assertDiracFailsWith(
        self.whiz._checkConsistency(
            Mock()),
        'Multiple instances of "sqrts=..GeV" detected, only one can be processed',
        self)

  def test_checkconsistency_nouserjob_fails5(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'sqrts=350 GeV\n model=SM\n model=Susy \ndecay_proc'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    assertDiracFailsWith(
        self.whiz._checkConsistency(
            Mock()),
        'Multiple instances of "model=..." detected, only one can be processed',
        self)

  def test_checkconsistency_nouserjob_fails6(self):
    self.whiz.version = '2.3.1'
    self.whiz.whizard2SinFile = 'sqrts=350 GeV\n model=SM\n model=Susy'
    self.whiz.eventType = 'ee -> ff'
    self.whiz._jobtype = 'notUser'
    self.whiz.numberOfEvents = 100
    assertDiracFailsWith(self.whiz._checkConsistency(Mock()), '"decay_proc" not found', self)

  def test_setProcessVariables(self):
    ret = self.whiz.setProcessVariables('  decay_proc  ')
    self.assertTrue(ret['OK'])
    self.assertEqual(self.whiz._decayProc, ['decay_proc'])

    ret = self.whiz.setProcessVariables(('  decay_proc  ', 'decay_another   '))
    self.assertTrue(ret['OK'])
    self.assertEqual(self.whiz._decayProc, ['decay_proc', 'decay_another'])

    ret = self.whiz.setProcessVariables({'  decay_proc  ': 'decay_another   '})
    self.assertFalse(ret['OK'])
    self.assertIn('Cannot handle', ret.get('Message'))

  def test_setIntegratedProcess_lfn(self):
    lfn = 'lfn:/vo/user/u/username/mytar.tar.gz'
    ret = self.whiz.setIntegratedProcess(lfn)
    self.assertTrue(ret['OK'])
    self.assertEqual('', self.whiz._integratedProcess)
    self.assertIn(lfn, self.whiz.inputSB)

  @parameterized.expand([(True, '3.1.4', S_OK({'tt': 'tt.tar.gz'})),
                         (False, '', None),  # no version defined
                         (False, '3.1.4', S_OK({'nottt': 'nottt.tar.gz'})),
                       ])
  def test_setIntegratedProcess_config(self, success, version, knownProc):
    proc = 'tt'
    self.whiz.version = version
    with patch.object(self.whiz._ops, 'getOptionsDict', new=Mock(return_value=knownProc)):
      ret = self.whiz.setIntegratedProcess(proc)
    self.assertEqual(ret['OK'], success)
    if success:
      self.assertEqual(proc, self.whiz._integratedProcess)


def runTests():
  """Runs our tests."""
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(Whizard2TestCase)
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)
  print(testResult)


if __name__ == '__main__':
  runTests()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Create Full Chain."""
from __future__ import print_function
from __future__ import absolute_import
from DIRAC.Core.Base import Script
Script.parseCommandLine()

from ILCDIRAC.Interfaces.API.DiracILC import DiracILC
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Whizard, Mokka, Marlin, OverlayInput
from ILCDIRAC.Interfaces.API.NewInterface.UserJob import UserJob

from DIRAC import exit as dexit

dirac = DiracILC()

# wh.setOutputFile("myfile.stdhep")

j = UserJob()

wh = Whizard(processlist=dirac.getProcessList())
wh.setEnergy(3000)
wh.setEvtType("ee_h_mumu")
wh.setNbEvts(1)
wh.setEnergy(3000)
params = {}
params['USERB1'] = 'F'
wh.setParameterDict(params)
wh.setModel("sm")
res = j.append(wh)
if not res['OK']:
  print(res['Message'])
  dexit(1)


mo = Mokka()
mo.getInputFromApp(wh)
mo.setVersion("0706P08")
mo.setSteeringFile("clic_ild_cdr.steer")
mo.setNbEvts(1)
mo.setOutputFile("somefile.slcio")
res = j.append(mo)
if not res['OK']:
  print(res['Message'])
  dexit(1)


ov = OverlayInput()
ov.setDetectorModel("CLIC_ILD_CDR")
ov.setBXOverlay(60)
ov.setGGToHadInt(3.2)
ov.setNbSigEvtsPerJob(1)
ov.setBkgEvtType("gghad")
res = j.append(ov)
if not res['OK']:
  print(res['Message'])
  dexit(1)


ma = Marlin()
ma.setVersion("v0111Prod")
ma.setSteeringFile("clic_ild_cdr_steering_overlay.xml")
ma.setGearFile("clic_ild_cdr.gear")
ma.getInputFromApp(mo)
ma.setDebug(True)
res = j.append(ma)
if not res['OK']:
  print(res['Message'])
  dexit(1)
# print appplication's attributes.
ma.listAttributes()

j.setName("test")
j.setOutputSandbox("*.log")

res = dirac.checkparams(j)
if not res['OK']:
  print(res['Message'])
  dexit(1)



#!/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Add software from CVMFS to the CS.

Give list of applications, init_script path, MokkaDBSlice, [Clic|ILD]ConfigPath (if set)


Options:

  -P, --Platform <value>            Platform ex. x86_64-slc5-gcc43-opt
  -A, --Applications <value>        Comma separated list of applications
  -V, --Version <value>             Version name
  -C, --Comment <value>             Comment
  -S, --Script <value>              Full path to initScript
  -B, --Base <value>                Path to Installation Base
  -O, --Config <value>              Path To [Clic|ILD]Config (if it is in ApplicationPath)
  -Q, --DBSlice <value>             Path to Mokka DB Slice
  --MarlinPandoraVersion <value>    Version of MarlinPandora, needed for pandora_calibration_scripts
  --PandoraAnalysisVersion <value>  Version of PandoraAnalysis, needed for pandora_calibration_scripts
  -N, --dry-run                     DryRun: do not commit to CS


:since: Feb 18, 2015
"""
from __future__ import absolute_import
import os
import subprocess

from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import gLogger, gConfig, S_OK, S_ERROR, exit as dexit

__RCSID__ = "$Id$"


class _Params(object):
  """Collection of Parameters set via CLI switches."""

  def __init__(self):
    self.version = ''
    self.platform = 'x86_64-slc5-gcc43-opt'
    self.comment = ''
    self.applicationSet = set()
    self.dbSliceLocation = ''
    self.initScriptLocation = ''
    self.basePath = ''
    self.configPath = ''
    self.dryRun = False
    self.pandoraAnalysisVersion = 'HEAD'
    self.marlinPandoraVersion = 'HEAD'

  def setVersion(self, optionValue):
    self.version = optionValue
    return S_OK()

  def setPlatform(self, optionValue):
    self.platform = optionValue
    return S_OK()

  def setName(self, optionValue):
    apps = optionValue.split(',')
    self.applicationSet = {_.strip() for _ in apps}
    return S_OK()

  def setComment(self, optionValue):
    self.comment = optionValue
    return S_OK()

  def setDBSlice(self, optionValue):
    self.dbSliceLocation = optionValue
    return S_OK()

  def setInitScript(self, optionValue):
    self.initScriptLocation = optionValue
    return S_OK()

  def setBasePath(self, optionValue):
    self.basePath = optionValue
    return S_OK()

  def setConfig(self, optionValue):
    self.configPath = optionValue
    return S_OK()

  def setDryrun(self, _):
    self.dryRun = True
    return S_OK()

  def _setPAV(self, val):
    self.pandoraAnalysisVersion = val
    return S_OK()

  def _setMPV(self, val):
    self.marlinPandoraVersion = val
    return S_OK()

  def checkConsistency(self):
    """Check if all necessary parameter were defined."""

    if not self.applicationSet:
      return S_ERROR("No applications have beend defined")

    if not self.version:
      return S_ERROR("Version must be given")

    appListLower = {_.lower() for _ in self.applicationSet}

    # if we have only config applications we do not need the initScript or the
    # basepath, just the config path
    if all(not app.endswith('config') for app in appListLower):

      if not self.initScriptLocation:
        return S_ERROR("Initscript location is not defined")

      if not self.basePath:
        return S_ERROR("BasePath is not defined")

    if 'mokka' in appListLower and not self.dbSliceLocation:
      return S_ERROR("Mokka in application list, but not dbSlice location given")

    # if we have a config package we need the configPath
    if any(app.endswith('config') for app in appListLower) and not self.configPath:
      return S_ERROR("Config in application list, but no location given")

    for val in (self.initScriptLocation, self.basePath, self.dbSliceLocation, self.configPath):
      if val and not os.path.exists(val):
        gLogger.error("Cannot find this path:", val)
        return S_ERROR("CVMFS not mounted, or path is misstyped")

    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch("P:", "Platform=", "Platform ex. %s" % self.platform, self.setPlatform)
    Script.registerSwitch("A:", "Applications=", "Comma separated list of applications", self.setName)
    Script.registerSwitch("V:", "Version=", "Version name", self.setVersion)
    Script.registerSwitch("C:", "Comment=", "Comment", self.setComment)
    Script.registerSwitch("S:", "Script=", "Full path to initScript", self.setInitScript)
    Script.registerSwitch("B:", "Base=", "Path to Installation Base", self.setBasePath)

    Script.registerSwitch("O:", "Config=", "Path To [Clic|ILD]Config (if it is in ApplicationPath)", self.setConfig)

    Script.registerSwitch("Q:", "DBSlice=", "Path to Mokka DB Slice", self.setDBSlice)
    Script.registerSwitch('', 'MarlinPandoraVersion=', 'Version of MarlinPandora, needed for '
                          'pandora_calibration_scripts', self._setMPV)
    Script.registerSwitch('', 'PandoraAnalysisVersion=', 'Version of PandoraAnalysis, needed for '
                          'pandora_calibration_scripts', self._setPAV)

    Script.registerSwitch("N", "dry-run", "DryRun: do not commit to CS", self.dryRun)

    Script.setUsageMessage('\n'.join([__doc__.split('\n')[1],
                                         '\nUsage:',
                                         '  %s [option|cfgfile] ...\n' % Script.scriptName]))


class CVMFSAdder(object):
  """Container for all the objects and functions to add software to ILCDirac."""

  def __init__(self, cliParams):
    from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
    self.modifiedCS = False
    self.softSec = "/Operations/Defaults/AvailableTarBalls"
    self.mailadress = 'ilc-dirac@cern.ch'
    self.cliParams = cliParams
    self.parameter = dict(softSec=self.softSec,
                           platform=cliParams.platform,
                           version=cliParams.version,
                           basepath=cliParams.basePath,
                           initsctipt=cliParams.initScriptLocation,
                           marlinPandoraVersion=cliParams.marlinPandoraVersion,
                           pandoraAnalysisVersion=cliParams.pandoraAnalysisVersion,
                         )
    self.applications = cliParams.applicationSet
    self.detmodels = {}
    self.csAPI = CSAPI()

  def findDDSimDetectorModels(self):
    """find all detector models in lcgeo and fill the self.detmodels dictionary with Detmodel as key and path as value.

    :returns: None
    """
    for root, dirs, _files in os.walk(os.path.join(self.parameter["basepath"], "lcgeo")):
      for direct in dirs:
        if root.endswith("compact"):
          # the main xml file must have the same name as the folder
          xmlPath = os.path.join(root, direct, direct + ".xml")
          if os.path.exists(xmlPath):
            self.detmodels[direct] = xmlPath

    # for key4hep
    for root, dirs, _files in os.walk(self.getFromInitScript('lcgeo_DIR')):
      for direct in dirs:
        if root.endswith("compact"):
          # the main xml file must have the same name as the folder
          xmlPath = os.path.join(root, direct, direct + ".xml")
          if os.path.exists(xmlPath):
            self.detmodels[direct] = xmlPath


  def findFCCDetectorModels(self):
    """find all detector models in fccdetectors and fill the self.detmodels dictionary with Detmodel as
key and path as value.

    :returns: None
    """
    detectorBasePath = self.getFromInitScript('FCCDETECTORS')
    gLogger.notice('Looking for detectors in:', repr(detectorBasePath))
    for root, dirs, files in os.walk(detectorBasePath):
      if root.endswith('compact'):
        # the main xml file is called FCCee_DectMaster.xml
        mainFile = [theFile for theFile in files if theFile.endswith('DectMaster.xml')]
        if len(mainFile) == 1:
          mainFile = mainFile[0]
        elif not mainFile:
          gLogger.error('Found no Master file in', root)
          continue
        else:
          gLogger.error('Found more than one masterFile in', (root, mainFile))
          continue
        # look for DetModelName/compact/mainFile
        detModel = root.rsplit('/', 2)[-2]
        xmlPath = os.path.join(root, mainFile)
        if os.path.exists(xmlPath):
          gLogger.notice('Found', xmlPath)
          self.detmodels[detModel] = xmlPath
    gLogger.notice('Found %d detectors' % len(self.detmodels))

  def checkConsistency(self):
    """checks if platform is defined, application exists, etc."""
    gLogger.notice("Checking consistency")
    av_platforms = gConfig.getSections(self.softSec, [])
    if av_platforms['OK']:
      if not self.parameter['platform'] in av_platforms['Value']:
        gLogger.error("Platform %s unknown, available are %s." %
                      (self.parameter['platform'], ", ".join(av_platforms['Value'])))
        gLogger.error("If yours is missing, add it in CS")
        return S_ERROR()
    else:
      gLogger.error("Could not find all platforms available in CS", av_platforms ['Message'])
      return S_ERROR()

    for application in self.applications:
      av_apps = gConfig.getSections("%(softSec)s/%(platform)s/" % self.parameter + str(application), [])
      if not av_apps['OK']:
        gLogger.error("Could not find this application in the CS: '%s'" % application)
        gLogger.error("Add its section to the CS, if it is missing")
        return S_ERROR()

    gLogger.notice("All OK, continuing...")
    return S_OK()

  def commitToCS(self):
    """write changes to the CS to the server."""
    if self.modifiedCS and not self.cliParams.dryRun:
      gLogger.notice("Commiting changes to the CS")
      result = self.csAPI.commit()
      if not result['OK']:
        gLogger.error('Commit failed with message = %s' % (result['Message']))
        return S_ERROR("Failed to commit to CS")
      gLogger.info('Successfully committed changes to CS')
    else:
      gLogger.info('No modifications to CS required')
    return S_OK()

  def addAllToCS(self):
    """add all the applications to the CS, take care of special cases (mokka, ildconfig, ddsim,...)"""
    from ILCDIRAC.ILCTransformationSystem.Utilities.ReleaseHelper import insertCSSection

    for application in self.applications:
      csParameter = dict(CVMFSEnvScript=self.cliParams.initScriptLocation,
                          CVMFSPath=self.parameter['basepath']
                        )

      if application == 'mokka':
        csParameter['CVMFSDBSlice'] = self.cliParams.dbSliceLocation

      if application == 'pandora_calibration_scripts':
        csParameter['MarlinPandora'] = os.path.join(self.parameter['basepath'], 'MarlinPandora',
                                                    self.parameter['marlinPandoraVersion'], 'scripts')
        csParameter['PandoraAnalysis'] = os.path.join(self.parameter['basepath'], 'PandoraAnalysis',
                                                      self.parameter['pandoraAnalysisVersion'], 'bin')
        if not os.path.exists(csParameter['PandoraAnalysis']):
          return S_ERROR('PandoraAnalysis folder %r not found' % csParameter['PandoraAnalysis'])
        if not os.path.exists(csParameter['MarlinPandora']):
          return S_ERROR('MarlinPandora folder %r not found' % csParameter['MarlinPandora'])
        del csParameter['CVMFSPath']

      if application == 'ddsim':
        self.findFCCDetectorModels()
        self.findDDSimDetectorModels()

        csPathModels = "Operations/Defaults/DDSimDetectorModels"
        csModels = {self.parameter["version"]: self.detmodels}
        insertCSSection(self.csAPI, csPathModels, csModels)
        self.modifiedCS = True

      if application == 'gaudiapp':
        self.findFCCDetectorModels()
        self.findDDSimDetectorModels()

        csPathModels = "Operations/Defaults/DDSimDetectorModels"
        csModels = {self.parameter["version"]: self.detmodels}
        insertCSSection(self.csAPI, csPathModels, csModels)
        self.modifiedCS = True

      elif application.endswith('config'):
        del csParameter['CVMFSEnvScript']
        csParameter['CVMFSPath'] = self.cliParams.configPath
        if self.cliParams.dbSliceLocation:
          csParameter['CVMFSDBSlice'] = self.cliParams.dbSliceLocation

      resInsert = self.insertApplicationToCS(application, csParameter)
      if not resInsert['OK']:
        return resInsert

    return S_OK()

  def insertApplicationToCS(self, name, csParameter):
    """add given application found via CVMFS to the CS."""

    pars = dict(self.parameter)
    pars['name'] = name

    gLogger.notice("%(name)s: Adding version %(version)s to the CS" % pars)

    existingVersions = gConfig.getSections("%(softSec)s/%(platform)s/%(name)s" % pars, [])
    if not existingVersions['OK']:
      gLogger.error("Could not find all versions available in CS: %s" % existingVersions['Message'])
      dexit(255)
    if pars['version'] in existingVersions['Value']:
      gLogger.always('Application %s %s for %s already in CS, nothing to do' % (name.lower(),
                                                                                pars['version'],
                                                                                pars['platform']))
      return S_OK()

    csPath = self.softSec + ("/%(platform)s/%(name)s/%(version)s/" % pars)
    for par, val in csParameter.items():
      gLogger.notice("Add: %s = %s" % (csPath + par, val))
      result = self.csAPI.setOption(csPath + par, val)
      if result['OK']:
        self.modifiedCS = True
      else:
        gLogger.error("Failure to add to CS", result['Message'])
        return S_ERROR("")

    return S_OK()

  def addSoftware(self):
    """run all the steps to add software to grid and CS."""

    resAdd = self.addAllToCS()
    if not resAdd['OK']:
      return resAdd

    resCommit = self.commitToCS()
    if not resCommit['OK']:
      return resCommit

    return S_OK()

  def getFromInitScript(self, variable):
    """Return value of the variable as set by the init script."""
    command = "source %s > /dev/null; echo $%s" % (self.cliParams.initScriptLocation, variable)
    variableValue = subprocess.check_output(command, shell=True).strip()
    return variableValue

@DIRACScript()
def main():
  """uploads, registers, and sends email about new software package."""
  cliParams = _Params()
  cliParams.registerSwitches()
  Script.parseCommandLine(ignoreErrors=True)

  consistent = cliParams.checkConsistency()
  if not consistent['OK']:
    gLogger.error("Error checking consistency:", consistent['Message'])
    Script.showHelp()
    dexit(2)

  softAdder = CVMFSAdder(cliParams)
  resCheck = softAdder.checkConsistency()

  if not resCheck['OK']:
    Script.showHelp()
    dexit(2)

  softAdder.addSoftware()

  gLogger.notice("All done!")
  dexit(0)


if __name__ == "__main__":
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""This application is used to obtain the host information.

It has no input/output, only the log file matters
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from DIRAC import S_OK, S_ERROR


class CheckWNs(LCApplication):
  """Small utility to probe a worker node: list the machine's properties, the sharedarea, and check if CVMFS is present."""

  def __init__(self, paramdict=None):
    super(CheckWNs, self).__init__(paramdict)
    self._modulename = "AnalyseWN"
    self.appname = 'analysewns'
    self._moduledescription = 'Analyse the WN on which this app runs'
    self.Version = "1"
    self.accountInProduction = False

  def _applicationModule(self):
    m1 = self._createModuleDefinition()
    return m1

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    if not res1["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    if not res1["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):
    """Checks that all needed parameters are set."""
    return S_OK()
